"""
Adapater to connect public uuid queue functionality to a specific queue type
"""
import time

from .queues.base_queue import (
    BaseQueueClient,
    BASE_QUEUE_TYPE,
)
from .queues.redis_queues import (
    REDIS_LIST,
    REDIS_LIST_PIPE,
    REDIS_SET,
    REDIS_SET_PIPE,
    REDIS_SET_PIPE_EXEC,
    RedisClient,
)


def _get_combined_uuids_gen(batch_by, uuid_len, max_value_size, uuids):
    combo_uuid = ''
    combine_count = 0
    combo_uuid_size = 0
    for uuid in uuids:
        new_size = len(uuid)
        if not new_size == uuid_len:
            raise ValueError(
                '_get_combined_uuids_gen: uuids are not correct length.'
                ' %d != %d' % (new_size, uuid_len)
            )
        to_be_size = combo_uuid_size + new_size
        if combine_count < batch_by and to_be_size <= max_value_size:
            combine_count += 1
            combo_uuid_size = to_be_size
            combo_uuid += uuid
        else:
            yield combo_uuid
            combine_count = 1
            combo_uuid_size = new_size
            combo_uuid = uuid
    if combo_uuid:
        yield combo_uuid


def _get_uncombined_uuids(uuid_len, combo_uuid):
    uuids = []
    uuid_count = len(combo_uuid) // uuid_len
    for ind in range(0, uuid_count):
        start = uuid_len * ind
        uuid = combo_uuid[start:start  + uuid_len]
        uuids.append(uuid)
    return uuids


class QueueTypes(object):
    '''
    Queue Type Manager
    '''
    QUEUES = [
        (BASE_QUEUE_TYPE, BaseQueueClient),
        (REDIS_LIST, RedisClient),
        (REDIS_LIST_PIPE, RedisClient),
        (REDIS_SET, RedisClient),
        (REDIS_SET_PIPE, RedisClient),
        (REDIS_SET_PIPE_EXEC, RedisClient),
    ]

    @classmethod
    def check_queue_type(cls, queue_type):
        '''Verify Queue Type'''
        queues = cls.get_all()
        if queue_type in queues:
            return True
        return False

    @classmethod
    def get_all(cls):
        '''Return all available queue types'''
        values = []
        for queue_type, _ in cls.QUEUES:
            values.append(queue_type)
        return values

    @classmethod
    def get_queue_client_class(cls, given_queue_type):
        '''Get the client for given queue type'''
        for queue_type, queue_client_class in cls.QUEUES:
            if queue_type == given_queue_type:
                return queue_client_class
        return None


class QueueAdapter(object):
    """
    Indexer to Queue Adapter / Manager
    """
    def __init__(self, queue_name, queue_type, queue_options):
        self._queue_type = queue_type
        self._queue_options = queue_options
        self._start_us = int(time.time() * 1000000)
        self.queue_id = str(self._start_us)
        self._queue = self._get_queue(queue_name)
        if self._queue:
            self._queue_name = self._queue.queue_name


    # Errors
    def _has_errors(self):
        """Check if index session errors in queue meta through queue"""
        return self._queue.has_errors()

    def pop_errors(self):
        """Pop errors from queue meta through queue"""
        return self._queue.pop_errors()

    # Worker
    def _get_worker_id(self):
        worker_conn_cnt = self._queue.get_worker_conn_count()
        time_id = self._start_us + worker_conn_cnt + 1
        return str(time_id)

    def _get_worker_queue(self):
        """Queue to create workers with"""
        if self._queue.queue_type == BASE_QUEUE_TYPE:
            # For base queue types,
            # server and worker queue point to same obj
            worker_queue = self._queue
        else:
            worker_queue = self._get_queue(self._queue_name, is_worker=True)
        return worker_queue

    def get_worker(self):
        """Return the associated worker"""
        worker_id = self._get_worker_id()
        # Assume that work conn id does not already exist
        self._queue.add_worker_conn(worker_id)
        worker_queue = self._get_worker_queue()
        return WorkerAdapter(
            self._queue_name,
            self._queue_options,
            worker_id,
            worker_queue,
        )

    def update_worker_conn(self, worker_id, uuid_cnt, get_cnt):
        """Update worker connection in queue meta through queue"""
        self._queue.update_worker_conn(worker_id, uuid_cnt, get_cnt)

    # Uuids
    def has_uuids(self, errs_cnt=0):  # pylint: disable=unused-argument
        '''Are there uuids in the queue'''
        return self._queue.has_uuids(errs_cnt=errs_cnt)

    def _load_uuids(self, uuids): #
        """
        Call load uuids on specific queue.

        * Handles combining uuids if configured
        """
        failed_batches = []
        bytes_added = 0
        success_cnt = 0
        batch_size = self._queue_options['batch_size']
        uuid_len = self._queue_options['uuid_len']
        if batch_size == 1:
            bytes_added, failed_batches = self._queue.load_uuids(uuids)
        else:
            combined_uuids_gen = _get_combined_uuids_gen(
                batch_size,
                uuid_len,
                self._queue.max_value_size,
                uuids,
            )
            bytes_added, failed_batches = self._queue.load_uuids(
                combined_uuids_gen
            )
        success_cnt = bytes_added//uuid_len
        if failed_batches:
            # ToDo: Handle failed uuids
            print(
                'UuidQueue adapater, handle failed uuids on load: %d batches'
                ' failed' % len(failed_batches)
            )
        return success_cnt

    def load_uuids(self, uuids):
        '''
        Wrapper for loading uuids

        * Checks if indexing already running, one load per indexing session
        * Handles meta data
        '''
        if self.is_indexing():
            return None
        success_cnt = self._load_uuids(uuids)
        self._queue.update_uuid_count(success_cnt)
        return success_cnt

    # Run
    def is_indexing(self, errs_cnt=0):
        '''Is an indexing process currently running'''
        if self.has_uuids(errs_cnt=errs_cnt) or self._has_errors():
            return True
        worker_conns = self._queue.get_worker_conns()
        for worker_conn in worker_conns.values():
            if int(worker_conn['uuid_cnt']):
                return True
        return False

    # Queue Client
    def _get_queue(self, queue_name, is_worker=False):
        client_class = QueueTypes.get_queue_client_class(self._queue_type)
        if client_class:
            client = client_class(self._queue_options)
            return client.get_queue(
                queue_name,
                self._queue_type,
                is_worker=is_worker,
            )
        return None

    def close_indexing(self):
        '''Close indexing sessions'''
        self._queue.close_indexing()


class WorkerAdapter(object):
    '''
    UuidQueueWorker to adapt to all uuid queue types

    * For in memory base queue,
    the queue arg in init should be the same object as the server
    '''
    def __init__(self, queue_name, queue_options, worker_id, queue):
        self._queue_name = queue_name
        self._queue_options = queue_options
        self.worker_id = worker_id
        self._queue = queue
        self.is_running = False
        # Worker conn info
        self.get_cnt = 0
        self.uuid_cnt = 0
        self.processes = self._queue_options['processes']
        self.chunk_size = self._queue_options['chunk_size']

    # Uuids
    @staticmethod
    def _get_uncombined_uuids(uuid_len, combo_uuid):
        """Wrapper to call top level function"""
        return _get_uncombined_uuids(uuid_len, combo_uuid)

    def _get_uuids(self):
        '''The only way to get uuids from queue'''
        uuids = []
        batch_size = self._queue_options['batch_size']
        get_size = self._queue_options['get_size']
        get_count = get_size // batch_size + 1
        if batch_size == 1:
            uuids = self._queue.get_uuids(1)
        else:
            combined_uuids_list = self._queue.get_uuids(get_count)
            for combined_uuids in combined_uuids_list:
                uncombined_uuids = self._get_uncombined_uuids(
                    self._queue_options['uuid_len'],
                    combined_uuids,
                )
                uuids.extend(uncombined_uuids)
        return uuids

    def get_uuids(self, get_all=False):
        '''Get uuids and update queue meta data'''
        self.get_cnt += 1
        uuids = []
        if self.uuid_cnt == 0:
            if get_all:
                tmp_uuids = self._get_uuids()
                while tmp_uuids:
                    uuids.extend(tmp_uuids)
                    tmp_uuids = self._get_uuids()
            else:
                uuids = self._get_uuids()
            self.uuid_cnt = len(uuids)
            self._queue.update_uuid_count(-1 * self.uuid_cnt)
        self._queue.update_worker_conn(
            self.worker_id,
            self.uuid_cnt,
            self.get_cnt,
        )
        return uuids

    # Run
    def update_finished(self, batch_results):
        '''Update server with batch results'''
        self._queue.update_success_count(batch_results['successes'])
        self._queue.update_errors_count(len(batch_results['errors']))
        msg = self._queue.update_finished(self.worker_id, batch_results)
        if msg == 'Okay':
            self.uuid_cnt = 0
            msg = None
        else:
            msg = 'Update finished could not reset worker: %s' % msg
        return msg



"""Tests the uuid queue base queue with a mocked indexer"""
from unittest import (
    TestCase,
    mock,
)

import pytest

from snovault.elasticsearch.indexer import Indexer
from snovault.elasticsearch.uuid_queue import QueueAdapter
from snovault.elasticsearch.uuid_queue.adapter_queue import WorkerAdapter
from snovault.elasticsearch.uuid_queue.queues.redis_queues import (
    RedisQueue,
    REDIS_LIST,
)

from .test_indexer_simple import (
    _get_uuids,
    MockRegistry,
    MockRequest,
)

TEST_QUEUE_NAME = 'testindxQ'
TEST_REDIS_DB = 3


@pytest.mark.skipif(True, reason="Redis is not install by default")
class TestIndexerRedisQueue(TestCase):
    """Test Indexer in indexer.py with base queue set"""
    @classmethod
    def setUpClass(cls):
        # Request
        cls.embed_wait = 0.01
        # Server Objedcts
        cls.invalidated = _get_uuids(2500)
        cls.xmin = None
        cls.snapshot_id = 'Some-Snapshot-Id'
        cls.restart = False
        # Indexer
        batch_size = 100
        processes = 1
        registry = MockRegistry(
            batch_size,
            processes=processes,
            queue_name=TEST_QUEUE_NAME,
            queue_host='localhost',
            queue_port=6379,
            queue_db=TEST_REDIS_DB,
        )
        queue_type = REDIS_LIST
        registry.settings['queue_type'] = queue_type
        registry['available_queues'].append(queue_type)
        registry['UuidQueue'] = QueueAdapter
        cls.indexer = Indexer(registry)

    @classmethod
    def tearDownClass(cls):
        # pylint: disable=protected-access
        q_srv_meta = cls.indexer.queue_server._queue._qmeta
        for key in q_srv_meta._client.keys(TEST_QUEUE_NAME + '*'):
            q_srv_meta._client.delete(key)

    def tearDown(self):
        """Reset the queue and queue meta data"""
        # pylint: disable=protected-access
        q_srv_meta = self.indexer.queue_server._queue._qmeta
        q_srv_meta._client.set(q_srv_meta._key_addedcount, 0)
        q_srv_meta._client.set(q_srv_meta._key_successescount, 0)
        q_srv_meta._client.set(q_srv_meta._key_addedcount, 0)
        self.indexer.queue_worker.uuid_cnt = 0
        self.indexer.queue_worker.get_cnt = 0
        self.indexer.queue_worker.is_running = False

    def test_init(self):
        """Test Indexer Init"""
        q_srv = self.indexer.queue_server
        q_wrk = self.indexer.queue_worker
        self.assertIsInstance(q_srv, QueueAdapter)
        self.assertIsInstance(q_wrk, WorkerAdapter)
        # pylint: disable=protected-access
        self.assertIsInstance(q_srv._queue, RedisQueue)
        self.assertIsInstance(q_wrk._queue, RedisQueue)
        self.assertTrue(q_srv._queue is not q_wrk._queue)

    def test_serve_objects_no_uuids(self):
        """Test serve objects with no_uuids"""
        expected_err_msg = (
            'Cannot initialize indexing process: '
            'No uuids given to Indexer.serve_objects'
        )
        request = MockRequest(embed_wait=self.embed_wait)
        uuids = []
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])

    def test_serve_objects_noload(self):
        """Test serve objects when fails to load any uuids"""
        uuids = self.invalidated
        org_load_uuids = self.indexer.queue_server.load_uuids
        self.indexer.queue_server.load_uuids = mock.MagicMock(return_value=None)
        expected_err_msg = 'Uuids given to Indexer.serve_objects failed to load'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.load_uuids = org_load_uuids

    def test_serve_objects_partload(self):
        """Test serve objects when fails to load all uuids"""
        uuids_loaded_cnt = 1
        uuids = self.invalidated
        org_load_uuids = self.indexer.queue_server.load_uuids
        self.indexer.queue_server.load_uuids = mock.MagicMock(
            return_value=uuids_loaded_cnt
        )
        expected_err_msg = (
            'Uuids given to Indexer.serve_objects '
            'failed to all load. {} of {} only'.format(
                uuids_loaded_cnt,
                len(uuids),
            )
        )
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.load_uuids = org_load_uuids

    def test_serve_objects_isindexing(self):
        """Test serve objects when already indexing"""
        uuids = self.invalidated
        org_is_indexing = self.indexer.queue_server.is_indexing
        self.indexer.queue_server.is_indexing = mock.MagicMock(return_value=True)
        expected_err_msg = 'Cannot initialize indexing process: Already Indexing'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_server.is_indexing = org_is_indexing

    def test_serve_objects_noworker(self):
        """
        Test serve objects when no worker.  Should timeout for simple queue

        Since worker cannot finish we manually reset
            - self.indexer.queue_server._uuids = []
        """
        uuids = self.invalidated
        org_queue_worker = self.indexer.queue_worker
        self.indexer.queue_worker = None
        expected_err_msg = 'Indexer sleep timeout'
        request = MockRequest(embed_wait=self.embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=1,
        )
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])
        self.indexer.queue_worker = org_queue_worker

    def test_serve_objects_timeout(self):
        """
        Test serve objects timeout

        Since worker cannot finish we manually reset
            - self.indexer.queue_server._uuids = []
            - serve objects can only timeout after worker finishes
        """
        uuids = self.invalidated
        embed_wait = 0.01
        expected_err_msg = 'Indexer sleep timeout'
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=0.1,
        )
        # pylint: disable=protected-access
        self.indexer.queue_server._uuids = []
        self.assertEqual(err_msg, expected_err_msg)
        self.assertListEqual(errors, [])

    def test_serve_objects_batcherrs(self):
        """
        Test serve objects returns batch errors
        """
        uuids = list(self.invalidated)[0:100]
        embed_wait = 0.01
        return_errors = ['e1', 'e2', 'e3']
        org_pop_errors = self.indexer.queue_server.pop_errors
        self.indexer.queue_server.pop_errors = mock.MagicMock(
            return_value=return_errors
        )
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=10,
        )
        self.indexer.queue_server.pop_errors = org_pop_errors
        # pylint: disable=protected-access
        self.indexer.queue_server._uuids = []
        self.assertIsNone(err_msg)
        errors.sort()
        return_errors.sort()
        self.assertListEqual(errors, return_errors)

    def test_serve_objects_run(self):
        """
        Test serve objects run
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        # pylint: disable=protected-access
        self.indexer.queue_worker._queue_options['batch_size'] = batch_size
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertListEqual(errors, [])
        self.assertIsNone(err_msg)

    def test_serve_objects_runerrs(self):
        """
        Test serve objects run with errors
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        # pylint: disable=protected-access
        self.indexer.queue_worker._queue_options['batch_size'] = batch_size
        request = MockRequest(embed_wait=embed_wait)
        request.set_embed_errors(10)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(len(errors), 10)
        self.assertIsNone(err_msg)

    def test_serve_objects_twiceerrs(self):
        """
        Test serve objects running twice with first run having errors
        """
        uuids = self.invalidated
        embed_wait = 0.001
        batch_size = 1024
        # pylint: disable=protected-access
        self.indexer.queue_worker._queue_options['batch_size'] = batch_size
        # Error Run
        request = MockRequest(embed_wait=embed_wait)
        request.set_embed_errors(10)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertEqual(len(errors), 10)
        self.assertIsNone(err_msg)
        # Clean Run
        request = MockRequest(embed_wait=embed_wait)
        _, errors, err_msg = self.indexer.serve_objects(
            request,
            uuids,
            self.xmin,
            snapshot_id=self.snapshot_id,
            restart=self.restart,
            timeout=100,
        )
        self.assertListEqual(errors, [])
        self.assertIsNone(err_msg)



"""\
Move attachment blobs to S3.
This is provided as an example upgrade script, it was used by the ENCODE project to migrate
attachment BLOBs from postgres to s3
"""
import copy
import logging
import transaction
from hashlib import md5
from pyramid.paster import get_app
from pyramid.threadlocal import manager
from pyramid.testing import DummyRequest
from snovault.interfaces import (
    BLOBS,
    DBSESSION
)
from snovault.storage import (
    PropertySheet,
    RDBBlobStorage,
)

EPILOG = __doc__

logger = logging.getLogger(__name__)


def run(app):
    root = app.root_factory(app)
    dummy_request = DummyRequest(root=root, registry=app.registry, _stats={})
    manager.push({'request': dummy_request, 'registry': app.registry})
    session = app.registry[DBSESSION]()
    blob_storage = app.registry[BLOBS]
    rdb_blobs = RDBBlobStorage(app.registry[DBSESSION])

    for sheet in session.query(PropertySheet).filter(PropertySheet.name == 'downloads'):
        # Copy the properties so sqlalchemy realizes it changed after it's mutated
        properties = copy.deepcopy(sheet.properties)
        download_meta = properties['attachment']
        if 'bucket' not in download_meta:
            # Re-writing the blob while the S3BlobStorage is in use
            # will move it to S3.
            data = rdb_blobs.get_blob(download_meta)
            blob_id = download_meta.pop('blob_id')
            download_meta['md5sum'] = md5(data).hexdigest()
            blob_storage.store_blob(data, download_meta, blob_id=blob_id)
            sheet.properties = properties
            logger.info('Updated %s' % sheet.sid)


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="Move attachment blobs to S3", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('--abort', action='store_true', help="Rollback transaction")
    parser.add_argument('config_uri', help="path to configfile")
    args = parser.parse_args()

    logging.basicConfig()
    app = get_app(args.config_uri, args.app_name)
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('snovault').setLevel(logging.DEBUG)

    raised = False
    try:
        run(app)
    except:
        raised = True
        raise
    finally:
        if raised or args.abort:
            transaction.abort()
            logger.info('Rolled back.')
        else:
            transaction.commit()


if __name__ == '__main__':
    main()



import pytest
from ..loadxl import ORDER


@pytest.mark.parametrize('item_type', ORDER)
def test_create_mapping(registry, item_type):
    from snovault.elasticsearch.create_mapping import type_mapping
    from snovault import TYPES
    mapping = type_mapping(registry[TYPES], item_type)
    assert mapping



"""Test Redis Queue File"""
from unittest import (
    TestCase,
    mock,
)

from ..queues.redis_queues import (  # pylint: disable=unused-import
    REDIS_LIST,
    REDIS_LIST_PIPE,
    REDIS_SET,
    REDIS_SET_PIPE,
    REDIS_SET_PIPE_EXEC,
    RedisClient,
    RedisQueueMeta,
    RedisQueue,
    RedisListQueue,
    RedisListPipeQueue,
    RedisSetQueue,
    RedisSetPipeQueue,
    RedisSetPipeExecQueue,
)


MOCK_TIME = 123456.654321
REDIS_TEST_DB = 3
REDIS_QUEUES = [
    ('REDIS_LIST', REDIS_LIST),
    ('REDIS_LIST_PIPE', REDIS_LIST_PIPE),
    ('REDIS_SET', REDIS_SET),
    ('REDIS_SET_PIPE', REDIS_SET_PIPE),
    ('REDIS_SET_PIPE_EXEC', REDIS_SET_PIPE_EXEC),
]
REDIS_QUEUE_CLASSES = [
    (REDIS_LIST, RedisListQueue),
    (REDIS_LIST_PIPE, RedisListPipeQueue),
    (REDIS_SET, RedisSetQueue),
    (REDIS_SET_PIPE, RedisSetPipeQueue),
    (REDIS_SET_PIPE_EXEC, RedisSetPipeExecQueue),
]
DEFAULT_QUEUE_OPTS = {
    'queue_name': 'TestRedisClient',
    'db': REDIS_TEST_DB,
    'host': 'localhost',
    'port': 6379,
}


class TestRedisClient(TestCase):
    '''Test Redis Queue Client Class'''
    queue_options = DEFAULT_QUEUE_OPTS.copy()
    def test_get_queue(self):
        '''Test get_queue'''
        for queue_type, queue_class in REDIS_QUEUE_CLASSES:
            client = RedisClient(self.queue_options)
            queue = client.get_queue(self.queue_options['queue_name'], queue_type)
            self.assertIsInstance(queue, queue_class)

    def test_get_queue_bad_type(self):
        '''Test get_queue with bad queue type'''
        client = RedisClient(self.queue_options)
        try:
            queue_type = 'bad queue type'
            _ = client.get_queue(self.queue_options['queue_name'], queue_type)
        except ValueError:
            pass
        else:
            self.fail('ValueError expected')


class TestRedisQueueMeta(TestCase):
    '''Test Redis Queue Meta Class'''
    # pylint: disable=too-many-public-methods
    members = [
        ('_base_id', int),
        ('_client', RedisClient),
        ('queue_name', str),
        # run args
        ('_key_metabase', str),
        ('_key_runargs', str),
        ('_key_uuidcount', str),
        ('_key_addedcount', str),
        ('_key_errors', str),
        ('_key_uuid_errors', str),
        ('_key_errorscount', str),
        ('_key_successescount', str),
        ('_key_workers', str),
        ('_key_worker_conn', str),
        ('_key_worker_results', str),
    ]
    meth_func = [
        # Errors
        'add_errors',
        'has_errors',
        'pop_errors',
        # Worker
        'add_worker_conn',
        '_get_worker_conn',
        '_get_worker_ids',
        'get_worker_conns',
        'update_worker_conn',
        'save_work_results',
        # Uuids
        'get_uuid_count',
        'has_uuids',
        'update_uuid_count',
        'update_success_count',
        # Run Args
        '_setup_redis_keys',
        'set_args',
        'get_run_args',
        'set_run_args',
        # base - not used?
        'get_worker_conn_count',
        '_get_blank_worker',
        '_init_persistant_data', # not tested
        'get_server_restarts', # not tested
        'update_errors_count', # not tested
    ]

    @classmethod
    def start_redis_server(cls):
        '''Helper to start redis-server'''


    @classmethod
    def stop_redis_server(cls):
        '''Helper to stop redis-server'''


    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        cls.queue_type = REDIS_LIST
        cls.queue_options = DEFAULT_QUEUE_OPTS.copy()
        client = RedisClient(cls.queue_options)
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.queue_meta = RedisQueueMeta(
            cls.queue_options['queue_name'],
            client
        )
        cls.standard_errors = [
            {'uuid': str(cnt), 'error': 'e' + str(cnt)}
            for cnt in range(10)
        ]
        # pylint: disable=protected-access
        for key in cls.queue_meta._client.keys('*'):
            cls.queue_meta._client.delete(key)

    @classmethod
    def tearDownClass(cls):
        '''Clean up after running test class'''
        # pylint: disable=protected-access
        for key in cls.queue_meta._client.keys('*'):
            cls.queue_meta._client.delete(key)

    def setUp(self):
        '''Setup before each method call'''
        self.queue_meta.set_args()

    def tearDown(self):
        """Clean up after each method call"""
        # pylint: disable=protected-access
        self.queue_meta.pop_errors()
        self.queue_meta._client.delete(self.queue_meta._key_errorscount)
        worker_ids = self.queue_meta._client.lrange(
            self.queue_meta._key_workers, 0, -1
        )
        self.queue_meta._client.delete(self.queue_meta._key_workers)
        for worker_id in worker_ids:
            self.queue_meta._client.delete(
                self.queue_meta._key_worker_conn + ':' + worker_id
            )
        partial_keys = self.queue_meta._client.keys(
            self.queue_meta._key_worker_results + ':*'
        )
        for key in partial_keys:
            self.queue_meta._client.delete(key)
        self.queue_meta._client.delete(self.queue_meta._key_uuidcount)

    # private/protected
    def test_init_dir(self):
        '''Test RedisQueueMeta has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue_meta)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue_meta, item))
            self.assertIsInstance(
                getattr(self.queue_meta, item),
                item_type
            )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue_meta, item))
            self.assertTrue(
                str(type(getattr(self.queue_meta, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test basic queue meta init'''
        # pylint: disable=protected-access
        self.assertEqual(self.mock_time_us, self.queue_meta._base_id)
        errors = self.queue_meta.pop_errors()
        self.assertListEqual([], errors)
        self.assertEqual(0, self.queue_meta.get_uuid_count())
        worker_conns = self.queue_meta.get_worker_conns()
        self.assertDictEqual({}, worker_conns)

    # Errors
    def test_add_errors(self):
        '''Test add_errors'''
        errors_to_add = self.standard_errors.copy()
        expected_result = len(errors_to_add)
        result = self.queue_meta.add_errors(errors_to_add)
        self.assertEqual(expected_result, result)
        exp_sorted_errors = sorted(errors_to_add, key=lambda k: k['uuid'])
        errors = self.queue_meta.pop_errors()
        res_sorted_errors = sorted(errors, key=lambda k: k['uuid'])
        self.assertListEqual(exp_sorted_errors, res_sorted_errors)

    def test_add_errors_none(self):
        '''Test add_errors when no errors'''
        errors_to_add = []
        expected_result = len(errors_to_add)
        result = self.queue_meta.add_errors(errors_to_add)
        self.assertEqual(expected_result, result)
        errors = self.queue_meta.pop_errors()
        self.assertListEqual([], errors)

    def test_has_errors(self):
        '''Test has_errors'''
        errors_to_add = self.standard_errors.copy()
        self.queue_meta.add_errors(errors_to_add)
        self.assertTrue(self.queue_meta.has_errors())

    def test_has_errors_none(self):
        '''Test has_errors when no errors'''
        self.assertFalse(self.queue_meta.has_errors())

    def test_pop_errors(self):
        '''Test pop_errors'''
        errors_to_add = self.standard_errors.copy()
        self.queue_meta.add_errors(errors_to_add)
        errors = self.queue_meta.pop_errors()
        # pylint: disable=protected-access
        err_cnt = int(
            self.queue_meta._client.get(self.queue_meta._key_errorscount)
        )
        self.assertEqual(err_cnt, 0)
        exp_sorted_errors = sorted(errors_to_add, key=lambda k: k['uuid'])
        res_sorted_errors = sorted(errors, key=lambda k: k['uuid'])
        self.assertListEqual(exp_sorted_errors, res_sorted_errors)

    # Workers
    def test_add_worker_conn(self):
        '''Test add_worker_conn'''
        worker_id = 'test-worker-id-a'
        self.queue_meta.add_worker_conn(worker_id)
        # pylint: disable=protected-access
        worker_ids = self.queue_meta._client.lrange(
            self.queue_meta._key_workers, 0, -1
        )
        self.assertListEqual(worker_ids, [worker_id])
        for worker_id in worker_ids:
            worker_conn_key = self.queue_meta._key_worker_conn + ':' + worker_id
            worker_hash = self.queue_meta._client.hgetall(worker_conn_key)
            worker_hash['uuid_cnt'] = int(worker_hash['uuid_cnt'])
            worker_hash['get_cnt'] = int(worker_hash['get_cnt'])
            self.assertDictEqual(worker_hash, self.queue_meta._get_blank_worker())

    def test_get_worker_conn(self):
        '''Test _get_worker_conn'''
        worker_id = 'test-worker-id-b'
        self.queue_meta.add_worker_conn(worker_id)
        # pylint: disable=protected-access
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        worker_conn['uuid_cnt'] = int(worker_conn['uuid_cnt'])
        worker_conn['get_cnt'] = int(worker_conn['get_cnt'])
        self.assertDictEqual(worker_conn, self.queue_meta._get_blank_worker())

    def test_get_worker_conns(self):
        '''Test _get_worker_conns'''
        worker_ids = ['test-worker-id-1', 'test-worker-id-2', 'test-worker-id-3']
        for worker_id in worker_ids:
            self.queue_meta.add_worker_conn(worker_id)
        worker_conns = self.queue_meta.get_worker_conns()
        self.assertEqual(len(worker_conns), len(worker_ids))
        for worker_conn in worker_conns.values():
            self.assertTrue(worker_conn['uuid_cnt'], '0')
            self.assertTrue(worker_conn['get_cnt'], '0')

    def test_update_worker_conn(self):
        '''Test update_worker_conn'''
        worker_id = 'test-worker-id-c'
        self.queue_meta.add_worker_conn(worker_id)
        uuid_cnt = 5
        get_cnt = 6
        self.queue_meta.update_worker_conn(worker_id, uuid_cnt, get_cnt)
        # pylint: disable=protected-access
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        self.assertEqual(int(worker_conn['uuid_cnt']), uuid_cnt)
        self.assertEqual(int(worker_conn['get_cnt']), get_cnt)

    def test_update_worker_conn_none(self):
        '''Test update_worker_conn when no worker'''
        worker_id = 'test-worker-id-d'
        # pylint: disable=protected-access
        uuid_cnt = 5
        get_cnt = 6
        self.queue_meta.update_worker_conn(worker_id, uuid_cnt, get_cnt)
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        self.assertDictEqual(worker_conn, {})

    def test_save_work_results(self):
        '''Test save_work_results'''
        worker_id = 'test-worker-id-e'
        self.queue_meta.add_worker_conn(worker_id)
        given_results = {'errors': ['e3', 'e4'], 'successes': 5}
        results_count = 1
        self.queue_meta.save_work_results(worker_id, given_results.copy())
        # pylint: disable=protected-access
        worker_res_key = '{}:{}:{}'.format(
            self.queue_meta._key_worker_results,
            worker_id,
            results_count,
        )
        got_results = self.queue_meta._client.hgetall(worker_res_key)
        got_results['successes'] = int(got_results['successes'])
        got_results['errors'] = int(got_results['errors'])
        given_results['errors'] = len(given_results['errors'])
        self.assertDictEqual(got_results, given_results)
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        self.assertEqual(int(worker_conn['results_count']), 1)

    def test_save_work_results_nowrk(self):
        '''Test save_work_results when no worker'''
        worker_id = 'test-worker-id-e'
        given_results = {'errors': ['e1', 'e2'], 'successes': 4}
        results_count = 1
        self.queue_meta.save_work_results(worker_id, given_results.copy())
        # pylint: disable=protected-access
        worker_res_key = '{}:{}:{}'.format(
            self.queue_meta._key_worker_results,
            worker_id,
            results_count,
        )
        got_results = self.queue_meta._client.hgetall(worker_res_key)
        self.assertDictEqual(got_results, {})
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        self.assertFalse('results_count' in worker_conn)

    def test_save_work_results_twice(self):
        '''Test save_work_results twice'''
        worker_id = 'test-worker-id-e'
        self.queue_meta.add_worker_conn(worker_id)
        given_results_1 = {'errors': ['1e1', '1e2'], 'successes': 14}
        given_results_2 = {'errors': ['2e1', '2e2'], 'successes': 3}
        results_count = 2
        self.queue_meta.save_work_results(worker_id, given_results_1.copy())
        self.queue_meta.save_work_results(worker_id, given_results_2.copy())
        # pylint: disable=protected-access
        worker_res_key = '{}:{}:{}'.format(
            self.queue_meta._key_worker_results,
            worker_id,
            results_count,
        )
        got_results = self.queue_meta._client.hgetall(worker_res_key)
        got_results['successes'] = int(got_results['successes'])
        got_results['errors'] = int(got_results['errors'])
        given_results_2['errors'] = len(given_results_2['errors'])
        self.assertDictEqual(got_results, given_results_2)
        worker_conn = self.queue_meta._get_worker_conn(worker_id)
        self.assertEqual(int(worker_conn['results_count']), results_count)

    # Uuids
    def test_get_uuid_count(self):
        '''Test get_uuid_count'''
        uuid_count = 1047
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_uuidcount, uuid_count)
        self.assertTrue(self.queue_meta.get_uuid_count() == uuid_count)

    def test_get_uuid_count_empty(self):
        '''Test get_uuid_count with no uuids'''
        uuid_count = 0
        self.assertTrue(self.queue_meta.get_uuid_count() == uuid_count)

    def test_has_uuids_added(self):
        '''Test has_uuids when only added'''
        added_cnt = 11
        success_cnt = 0
        errors_cnt = 0
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, added_cnt)
        self.queue_meta._client.set(self.queue_meta._key_successescount, success_cnt)
        self.queue_meta._client.set(self.queue_meta._key_errorscount, errors_cnt)
        self.assertTrue(self.queue_meta.has_uuids())

    def test_has_uuids_success(self):
        '''Test has_uuids when all success'''
        added_cnt = 11
        success_cnt = added_cnt
        errors_cnt = 0
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, added_cnt)
        self.queue_meta._client.set(self.queue_meta._key_successescount, success_cnt)
        self.queue_meta._client.set(self.queue_meta._key_errorscount, errors_cnt)
        self.assertFalse(self.queue_meta.has_uuids())

    def test_has_uuids_errors(self):
        '''Test has_uuids when all errors'''
        added_cnt = 11
        success_cnt = 0
        errors_cnt = added_cnt
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, added_cnt)
        self.queue_meta._client.set(self.queue_meta._key_successescount, success_cnt)
        self.queue_meta._client.set(self.queue_meta._key_errorscount, errors_cnt)
        self.assertFalse(self.queue_meta.has_uuids())

    def test_has_uuids_parterr(self):
        '''Test has_uuids when some errors'''
        added_cnt = 11
        success_cnt = 5
        errors_cnt = 6
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, added_cnt)
        self.queue_meta._client.set(self.queue_meta._key_successescount, success_cnt)
        self.queue_meta._client.set(self.queue_meta._key_errorscount, errors_cnt)
        self.assertFalse(self.queue_meta.has_uuids())

    def test_has_uuids_parterr_run(self):
        '''Test has_uuids when some errors and still running'''
        added_cnt = 11
        success_cnt = 3
        errors_cnt = 5
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, added_cnt)
        self.queue_meta._client.set(self.queue_meta._key_successescount, success_cnt)
        self.queue_meta._client.set(self.queue_meta._key_errorscount, errors_cnt)
        self.assertTrue(self.queue_meta.has_uuids())

    def test_update_uuid_count(self):
        '''Test update_uuid_count'''
        update_cnt = 5
        self.queue_meta.update_uuid_count(update_cnt)
        # pylint: disable=protected-access
        self.assertTrue(
            int(self.queue_meta._client.get(self.queue_meta._key_uuidcount)) == 5
        )
        update_cnt = -3
        self.queue_meta.update_uuid_count(update_cnt)
        self.assertTrue(
            int(self.queue_meta._client.get(self.queue_meta._key_uuidcount)) == 2
        )
        update_cnt = 8
        self.queue_meta.update_uuid_count(update_cnt)
        self.assertTrue(
            int(self.queue_meta._client.get(self.queue_meta._key_uuidcount)) == 10
        )

    # Run Args
    @staticmethod
    def test_setup_redis_keys():
        '''
        Check all redis keys on class
        - Tested in init
        '''
        pass

    def test_set_args(self):
        '''test set_args'''
        # pylint: disable=protected-access
        self.queue_meta._client.set(self.queue_meta._key_addedcount, 10)
        errors_to_add = self.standard_errors.copy()
        _ = self.queue_meta.add_errors(errors_to_add)
        self.queue_meta._client.set(self.queue_meta._key_successescount, 10)
        worker_id = 'test-worker-id-a'
        self.queue_meta.add_worker_conn(worker_id)
        self.queue_meta.set_args()
        self.assertEqual(
            int(self.queue_meta._client.get(
                self.queue_meta._key_addedcount
            )),
            0
        )
        self.assertEqual(
            int(self.queue_meta._client.get(
                self.queue_meta._key_errorscount
            )),
            0
        )
        self.assertEqual(
            int(self.queue_meta._client.get(
                self.queue_meta._key_successescount
            )),
            0
        )
        self.assertEqual(
            len(self.queue_meta._client.lrange(
                self.queue_meta._key_workers, 0, -1
            )),
            1
        )

    def test_get_run_args(self):
        '''test get_run_args'''
        run_args = {
            'batch_by': 101,
            'restart': False,
            'snapshot_id': 'fake-snap-id',
            'uuid_len': 34,
            'xmin': 1234,
        }
        self.queue_meta.set_run_args(run_args)
        res_run_args = self.queue_meta.get_run_args()
        self.assertTrue('batch_by' in res_run_args)
        self.assertIsInstance(res_run_args['batch_by'], int)
        self.assertTrue('restart' in res_run_args)
        self.assertIsInstance(res_run_args['restart'], bool)
        self.assertTrue('uuid_len' in res_run_args)
        self.assertIsInstance(res_run_args['uuid_len'], int)
        self.assertTrue('xmin' in res_run_args)
        self.assertIsInstance(res_run_args['xmin'], int)
        self.assertDictEqual(res_run_args, run_args)

    @staticmethod
    def test_set_run_args():
        '''
        test set_run_args
        - done in test_get_run_args
        '''
        pass


class TestRedisQueue(TestCase):
    '''Test Redis Queue'''
    # pylint: disable=too-many-public-methods
    members = [
        ('add_str', str),
        ('get_str', str),
        ('len_str', str),
        ('max_value_size', int),
        ('queue_name', str),
        ('queue_type', str),
        ('_client', RedisClient),
        ('_qmeta', RedisQueueMeta),
    ]
    meth_func = [
        '_call_func',
        # Errors
        'add_errors',
        'has_errors',
        # Worker
        'add_worker_conn',
        'get_worker_conns',
        'get_worker_conn_count',
        'save_work_results',
        'update_worker_conn',
        # Uuids
        'has_uuids',
        '_get_uuid',
        'get_uuids',
        '_load_uuid',
        'load_uuids',
        'pop_errors',
        'update_uuid_count',
        # Run Args
        'update_finished',
        'close_indexing', # not tested
        'get_server_restarts', # not tested
        'update_errors_count', # not tested
        'update_success_count', # not tested
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        cls.queue_type = REDIS_LIST
        cls.queue_options = DEFAULT_QUEUE_OPTS.copy()
        cls.client = RedisClient(cls.queue_options)
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.queue = cls.client.get_queue(
            cls.queue_options['queue_name'],
            cls.queue_type
        )
        # pylint: disable=protected-access
        for key in cls.queue._qmeta._client.keys('*'):
            cls.queue._qmeta._client.delete(key)

    @classmethod
    def tearDownClass(cls):
        '''Clean up after running test class'''
        # pylint: disable=protected-access
        for key in cls.queue._qmeta._client.keys('*'):
            cls.queue._qmeta._client.delete(key)

    def setUp(self):
        '''Setup before each method call'''
        pass

    def tearDown(self):
        """Clean up after each method call"""
        # pylint: disable=protected-access
        self.queue._qmeta.pop_errors()
        self.queue._qmeta._client.delete(self.queue._qmeta._key_errorscount)
        worker_ids = self.queue._qmeta._client.lrange(
            self.queue._qmeta._key_workers, 0, -1
        )
        self.queue._qmeta._client.delete(self.queue._qmeta._key_workers)
        for worker_id in worker_ids:
            self.queue._qmeta._client.delete(
                self.queue._qmeta._key_worker_conn + ':' + worker_id
            )
        queue_meta_worker_keys = self.queue._qmeta._client.keys(
            self.queue._qmeta._key_worker_results + ':*'
        )
        for key in queue_meta_worker_keys:
            self.queue._qmeta._client.delete(key)
        self.queue._qmeta._client.delete(self.queue._qmeta._key_uuidcount)
        self.queue._client.delete(self.queue.queue_name)

    # private/protected
    def test_init_dir(self):
        '''Test RedisQueue has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue, item))
            self.assertIsInstance(
                getattr(self.queue, item),
                item_type
            )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue, item))
            self.assertTrue(
                str(type(getattr(self.queue, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test basic queue meta init'''
        self.assertEqual(
            self.queue.queue_name,
            self.queue_options['queue_name'] + '0'
        )
        # pylint: disable=protected-access
        self.assertTrue(self.queue._client is self.client)
        assert True

    def test_get_uuid(self):
        '''Test _get_uuid'''
        uuids_to_add = ['a', 'b', 'c']
        # pylint: disable=protected-access
        for uuid in uuids_to_add:
            self.queue._client.lpush(self.queue.queue_name, uuid)
        ret_uuids = self.queue._get_uuid()
        self.assertEqual(ret_uuids, uuids_to_add[-1])

    def test_load_uuid(self):
        '''Test _load_uuid'''
        uuids_to_add = ['a', 'b', 'c']
        res_bool = self.queue.load_uuids(uuids_to_add.copy())
        self.assertTrue(res_bool)
        # pylint: disable=protected-access
        ret_uuids = self.queue._client.lrange(self.queue.queue_name, 0, -1)
        ret_uuids.sort()
        uuids_to_add.sort()
        self.assertListEqual(ret_uuids, uuids_to_add)



import pytest

# Test for storage.keys

pytest_plugins = [
    'snovault.tests.serverfixtures',
    'snovault.tests.testappfixtures',
]

items = [
    {'name': 'one', 'alias': 'TEST1'},
    {'name': 'two', 'alias': 'TEST2'},
]

bad_items = [
    {'name': 'one', 'alias': 'BAD1'},
    {'name': 'bad', 'alias': 'TEST1'},
]


@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


@pytest.fixture(scope='session')
def app(DBSession, app_settings):
    from pyramid.config import Configurator
    from snovault import DBSESSION
    config = Configurator()
    config.registry.settings['local_storage_host'] = app_settings['local_storage_host']
    config.registry.settings['local_storage_port'] = app_settings['local_storage_port']
    config.registry.settings['local_storage_redis_index'] = app_settings['local_storage_redis_index']
    config.registry.settings['local_storage_timeout'] = app_settings['local_storage_timeout']
    config.registry.settings['local_tz'] = app_settings['local_tz']
    config.registry[DBSESSION] = DBSession
    config.include('snovault')
    config.include('.testing_key')
    return config.make_wsgi_app()


@pytest.fixture
def content(testapp):
    url = '/testing-keys/'
    for item in items:
        testapp.post_json(url, item, status=201)


@pytest.mark.parametrize('item', items)
def test_unique_key(testapp, content, item):
    url = '/testing-keys/' + item['alias']
    res = testapp.get(url).maybe_follow()
    assert res.json['name'] == item['name']


def test_keys_bad_items(testapp, content):
    url = '/testing-keys/'
    for item in bad_items:
        testapp.post_json(url, item, status=409)


def test_keys_update(testapp):
    url = '/testing-keys/'
    item = items[0]
    res = testapp.post_json(url, item, status=201)
    location = res.location
    new_item = {'name': 'new_one', 'alias': 'NEW1'}
    testapp.put_json(location, new_item, status=200)
    testapp.post_json(url, item, status=201)
    testapp.put_json(location, item, status=409)


def test_keys_conflict(testapp):
    url = '/testing-keys/'
    item = items[1]
    initial = testapp.get(url).json['@graph']
    testapp.post_json(url, item, status=201)
    posted = testapp.get(url).json['@graph']
    assert(len(initial)+1 == len(posted))
    conflict = testapp.post_json(url, item, status=409)
    assert(conflict.status_code == 409)
    conflicted = testapp.get(url).json['@graph']
    assert(len(posted) == len(conflicted))



from snovault.schema_utils import validate
import pytest


targets = [
    {'name': 'one', 'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f'},
]


@pytest.fixture
def content(testapp):
    url = '/testing-link-targets/'
    for item in targets:
        testapp.post_json(url, item, status=201)


def test_uniqueItems_validates_normalized_links(content, threadlocals):
    schema = {
        'uniqueItems': True,
        'items': {
            'linkTo': 'TestingLinkTarget',
        }
    }
    uuid = targets[0]['uuid']
    data = [
        uuid,
        '/testing-link-targets/{}'.format(uuid),
    ]
    validated, errors = validate(schema, data)
    assert len(errors) == 1
    assert (
        errors[0].message == "['{}', '{}'] has non-unique elements".format(
            uuid, uuid)
    )



from collections import defaultdict
from functools import reduce
from pyramid.decorator import reify
from .interfaces import (
    CALCULATED_PROPERTIES,
    TYPES,
)
from .schema_utils import combine_schemas


def includeme(config):
    registry = config.registry
    registry[TYPES] = TypesTool(registry)


def extract_schema_links(schema):
    if not schema:
        return
    for key, prop in schema['properties'].items():
        if 'items' in prop:
            prop = prop['items']
        if 'properties' in prop:
            for path in extract_schema_links(prop):
                yield (key,) + path
        elif 'linkTo' in prop:
            yield (key,)


class AbstractTypeInfo(object):
    factory = None

    def __init__(self, registry, name):
        self.types = registry[TYPES]
        self.name = name

    @reify
    def subtypes(self):
        return [
            ti.name for ti in self.types.by_item_type.values()
            if self.name in ([ti.name] + ti.base_types)
        ]

    @reify
    def schema(self):
        subschemas = (self.types[name].schema for name in self.subtypes)
        return reduce(combine_schemas, subschemas)


class TypeInfo(AbstractTypeInfo):
    def __init__(self, registry, item_type, factory):
        super(TypeInfo, self).__init__(registry, factory.__name__)
        self.registry = registry
        self.item_type = item_type
        self.factory = factory
        self.base_types = factory.base_types
        self.embedded = factory.embedded
        self.embedded_with_frame = factory.embedded_with_frame
        self.embedded_paths = factory.embedded_paths()

    @reify
    def calculated_properties(self):
        return self.registry[CALCULATED_PROPERTIES]

    @reify
    def schema_version(self):
        try:
            return self.factory.schema['properties']['schema_version']['default']
        except (KeyError, TypeError):
            return None

    @reify
    def schema_links(self):
        return sorted('.'.join(path) for path in extract_schema_links(self.factory.schema))

    @reify
    def schema_keys(self):
        if not self.factory.schema:
            return ()
        keys = defaultdict(list)
        for key, prop in self.factory.schema['properties'].items():
            uniqueKey = prop.get('items', prop).get('uniqueKey')
            if uniqueKey is True:
                uniqueKey = '%s:%s' % (self.factory.item_type, key)
            if uniqueKey is not None:
                keys[uniqueKey].append(key)
        return keys

    @reify
    def merged_back_rev(self):
        merged = {}
        types = [self.name] + self.base_types
        for name in reversed(types):
            back_rev = self.types.type_back_rev.get(name, ())
            merged.update(back_rev)
        return merged

    @reify
    def schema(self):
        props = self.calculated_properties.props_for(self.factory)
        schema = self.factory.schema or {'type': 'object', 'properties': {}}
        schema = schema.copy()
        schema['properties'] = schema['properties'].copy()
        for name, prop in props.items():
            if prop.schema is not None:
                schema['properties'][name] = prop.schema
        return schema

    @reify
    def schema_rev_links(self):
        revs = {}
        for key, prop in self.schema['properties'].items():
            if prop.get('notSubmittable'):
                continue
            linkFrom = prop.get('linkFrom', prop.get('items', {}).get('linkFrom'))
            if linkFrom is None:
                continue
            linkType, linkProp = linkFrom.split('.')
            revs[key] = linkType, linkProp
        return revs


class TypesTool(object):
    def __init__(self, registry):
        self.registry = registry
        self.by_item_type = {}
        self.abstract = {}
        self.type_back_rev = {}
        self.all = {}

    def register(self, factory):
        name = factory.__name__
        item_type = factory.item_type or name
        ti = TypeInfo(self.registry, item_type, factory)
        self.all[ti.item_type] = self.by_item_type[ti.item_type] = ti
        self.all[ti.name] = self.abstract[ti.name] = ti
        self.all[ti.factory] = ti
        for base in ti.base_types:
            self.register_abstract(base)

        # Calculate the reverse rev map
        for prop_name, spec in factory.rev.items():
            rev_type_name, rel = spec
            back = self.type_back_rev.setdefault(rev_type_name, {}).setdefault(rel, set())
            back.add((ti.name, prop_name))

        return ti

    def register_abstract(self, name):
        ti = self.abstract.get(name)
        if ti is None:
            ti = AbstractTypeInfo(self.registry, name)
            self.all[name] = self.abstract[name] = ti
        return ti

    def __contains__(self, name):
        return name in self.all

    def __getitem__(self, name):
        return self.all[name]



import pytest


@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


@pytest.fixture(scope='session')
def app_settings(wsgi_server_host_port, elasticsearch_server, postgresql_server, redis_server):
    from .testappfixtures import _app_settings
    settings = _app_settings.copy()
    settings['create_tables'] = True
    settings['persona.audiences'] = 'http://%s:%s' % wsgi_server_host_port
    settings['elasticsearch.server'] = elasticsearch_server
    settings['sqlalchemy.url'] = postgresql_server
    settings['collection_datastore'] = 'elasticsearch'
    settings['item_datastore'] = 'elasticsearch'
    settings['indexer'] = True
    settings['indexer.processes'] = 2
    settings['queue_type'] = 'Simple'
    settings['queue_server'] = True
    settings['queue_worker'] = True
    settings['queue_worker_processes'] = 2
    settings['queue_worker_chunk_size'] = 1024
    settings['queue_worker_batch_size'] = 2000
    return settings


@pytest.yield_fixture(scope='session')
def app(app_settings):
    from snovault import main
    app = main({}, **app_settings)

    yield app

    # Shutdown multiprocessing pool to close db conns.
    from snovault.elasticsearch import INDEXER
    app.registry[INDEXER].shutdown()

    from snovault import DBSESSION
    DBSession = app.registry[DBSESSION]
    # Dispose connections so postgres can tear down.
    DBSession.bind.pool.dispose()


@pytest.fixture(scope='session')
def DBSession(app):
    from snovault import DBSESSION
    return app.registry[DBSESSION]


@pytest.fixture(autouse=True)
def teardown(app, dbapi_conn):
    from snovault.elasticsearch import create_mapping
    create_mapping.run(app)
    cursor = dbapi_conn.cursor()
    cursor.execute("""TRUNCATE resources, transactions CASCADE;""")
    cursor.close()


@pytest.yield_fixture
def dbapi_conn(DBSession):
    connection = DBSession.bind.pool.unique_connection()
    connection.detach()
    conn = connection.connection
    conn.autocommit = True
    yield conn
    conn.close()


@pytest.yield_fixture
def listening_conn(dbapi_conn):
    cursor = dbapi_conn.cursor()
    cursor.execute("""LISTEN "snovault.transaction";""")
    yield dbapi_conn
    cursor.close()



import os
from snovault import DBSESSION
from contextlib import contextmanager
from multiprocessing import get_context
from multiprocessing.pool import Pool
from pyramid.decorator import reify
from pyramid.request import apply_request_extensions
from pyramid.threadlocal import (
    get_current_request,
    manager,
)
import atexit
import logging
import time
import transaction
from .indexer import (
    INDEXER,
    Indexer,
)
from .interfaces import (
    APP_FACTORY,
    ELASTIC_SEARCH,
)

log = logging.getLogger('snovault.elasticsearch.es_index_listener')


def includeme(config):
    if config.registry.settings.get('indexer_worker'):
        return
    processes = config.registry.settings.get('indexer.processes')
    try:
        processes = int(processes)
    except:
        processes = None
    if INDEXER not in config.registry:
        config.registry[INDEXER] = MPIndexer(config.registry, processes=processes)


# Running in subprocess

current_xmin_snapshot_id = None
app = None


def initializer(app_factory, settings):
    import signal
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    global app
    atexit.register(clear_snapshot)
    app = app_factory(settings, indexer_worker=True, create_tables=False)
    signal.signal(signal.SIGALRM, clear_snapshot)


def set_snapshot(xmin, snapshot_id):
    global current_xmin_snapshot_id
    if current_xmin_snapshot_id == (xmin, snapshot_id):
        return
    clear_snapshot()
    current_xmin_snapshot_id = (xmin, snapshot_id)

    while True:
        txn = transaction.begin()
        txn.doom()
        if snapshot_id is not None:
            txn.setExtendedInfo('snapshot_id', snapshot_id)
        session = app.registry[DBSESSION]()
        connection = session.connection()
        db_xmin = connection.execute(
            "SELECT txid_snapshot_xmin(txid_current_snapshot());").scalar()
        if db_xmin >= xmin:
            break
        transaction.abort()
        log.info('Waiting for xmin %r to reach %r', db_xmin, xmin)
        time.sleep(0.1)

    registry = app.registry
    request = app.request_factory.blank('/_indexing_pool')
    request.registry = registry
    request.datastore = 'database'
    apply_request_extensions(request)
    request.invoke_subrequest = app.invoke_subrequest
    request.root = app.root_factory(request)
    request._stats = {}
    manager.push({'request': request, 'registry': registry})


def clear_snapshot(signum=None, frame=None):
    global current_xmin_snapshot_id
    if current_xmin_snapshot_id is None:
        return
    transaction.abort()
    manager.pop()
    current_xmin_snapshot_id = None


@contextmanager
def snapshot(xmin, snapshot_id):
    import signal
    signal.alarm(0)
    set_snapshot(xmin, snapshot_id)
    yield
    signal.alarm(5)


def update_object_in_snapshot(args):
    uuid, xmin, snapshot_id, restart = args
    with snapshot(xmin, snapshot_id):
        request = get_current_request()
        encoded_es = request.registry[ELASTIC_SEARCH]
        map_info = {
            'start_time': time.time(),
            'end_time': None,
            'run_time': None,
            'pid':os.getpid(),
        }
        update_info = Indexer.update_object(
            encoded_es,
            request,
            uuid,
            xmin,
            restart=restart,
        )
        update_info['snapshot_id'] = snapshot_id
        map_info['end_time'] = time.time()
        map_info['run_time'] = map_info['end_time'] - map_info['start_time']
        update_info['map_info'] = map_info
        return update_info


# Running in main process

class MPIndexer(Indexer):
    maxtasks = 1  # pooled processes will exit and be replaced after this many tasks are completed.

    def __init__(self, registry, processes=None):
        super(MPIndexer, self).__init__(registry)
        self.initargs = (registry[APP_FACTORY], registry.settings,)

    @reify
    def pool(self):
        return Pool(
            processes=self.queue_worker.processes,
            initializer=initializer,
            initargs=self.initargs,
            maxtasksperchild=self.maxtasks,
            context=get_context('forkserver'),
        )

    def update_objects(
            self,
            request,
            uuids,
            xmin,
            snapshot_id=None,
            restart=False,
        ):
        # pylint: disable=too-many-arguments, unused-argument
        '''Run multiprocess indexing process on uuids'''
        # Ensure that we iterate over uuids in this thread not the pool task handler.
        processes = self.queue_worker.processes
        chunk_size = self.queue_worker.chunk_size
        uuid_count = len(uuids)
        chunkiness = int((uuid_count - 1) / processes) + 1
        if chunkiness > chunk_size:
            chunkiness = chunk_size
        tasks = [
            (uuid, xmin, snapshot_id, restart)
            for uuid in uuids
        ]
        errors = []
        update_infos = []
        start_time = time.time()
        try:
            for i, update_info in enumerate(
                    self.pool.imap_unordered(
                        update_object_in_snapshot,
                        tasks,
                        chunkiness)
                ):
                update_info['return_time'] = time.time()
                update_infos.append(update_info)
                error = update_info.get('error')
                if error is not None:
                    print('Error', error)
                    errors.append(error)
                if (i + 1) % 1000 == 0:
                    log.info('Indexing %d', i + 1)
        except:
            self.shutdown()
            raise
        return update_infos, errors

    def shutdown(self):
        if 'pool' in self.__dict__:
            self.pool.terminate()
            self.pool.join()
            del self.pool



import pytest

@pytest.fixture(scope='session')
def test_user():
    return {'username': 'user@test.com', 'password': 'test'}


def test_login_unknown_user(anontestapp, test_user):
    res = anontestapp.post_json('/login', test_user, status=403)

def test_invalid_create_user(testapp):
    url = '/users/'
    item = {
        'emailllll': 'blah@blah.com',
        'first_name': 'Persona',
        'last_name': 'Test User',
        'password' : 'password'
    }
    resp = testapp.post_json(url, item, status=422)
    print(resp)
    assert resp.json['errors'] is not None


def test_login_logout(testapp, anontestapp, app_settings, test_user):
    # Create a user with the test_users email
    url = '/users/'
    item = {
        'email': test_user['username'],
        'first_name': 'Persona',
        'last_name': 'Test User',
        'password' : test_user['password']
    }
    testapp.post_json(url, item, status=201)

    # Log in
    login = {'username' : item['email'], 'password': item['password']}
    res = anontestapp.post_json('/login', login, status=200)
    assert 'Set-Cookie' in res.headers
    assert res.json['auth.userid'] == login['username']

    # Log out
    res = anontestapp.get('/logout?redirect=false', status=200)
    assert 'Set-Cookie' in res.headers
    assert 'auth.userid' not in res.json

def test_impersonate_user(anontestapp, admin, submitter):
    res = anontestapp.post_json(
        '/impersonate-user', {'userid': submitter['email']},
        extra_environ={'REMOTE_USER': str(admin['email'])}, status=200)
    assert 'Set-Cookie' in res.headers
    assert res.json['auth.userid'] == submitter['email']
    res = anontestapp.get('/session-properties')
    assert res.json['auth.userid'] == submitter['email']



def test_render_error(anonhtmltestapp):
    res = anonhtmltestapp.get('/testing-render-error', status=500)
    assert res.body.startswith(b'<!DOCTYPE html>')


def test_render_error_multiple_times(anonhtmltestapp):
    anonhtmltestapp.get('/testing-render-error', status=500)
    res = anonhtmltestapp.get('/testing-render-error', status=500)
    assert res.body.startswith(b'<!DOCTYPE html>')


def test_render_error_then_success(anonhtmltestapp):
    anonhtmltestapp.get('/testing-render-error', status=500)
    res = anonhtmltestapp.get('/', status=200)
    assert res.body.startswith(b'<!DOCTYPE html>')



from pyramid.security import (
    Allow,
)
from pyramid.view import view_config
from snovault import (
    Item,
    calculated_property,
    collection,
)
from snowflakes.types.base import paths_filtered_by_status
from snovault.attachment import ItemWithAttachment
from snovault.util import Path
from snovault.elasticsearch.searches.configs import search_config


def includeme(config):
    config.scan(__name__)
    config.include('.testing_auditor')


@view_config(name='testing-user', request_method='GET')
def user(request):
    return {
        'authenticated_userid': request.authenticated_userid,
        'effective_principals': request.effective_principals,
    }


@view_config(name='testing-allowed', request_method='GET')
def allowed(context, request):
    from pyramid.security import (
        has_permission,
        principals_allowed_by_permission,
    )
    permission = request.params.get('permission', 'view')
    return {
        'has_permission': bool(has_permission(permission, context, request)),
        'principals_allowed_by_permission': principals_allowed_by_permission(context, permission),
    }


@collection(
    'testing-downloads',
    properties={
        'title': 'Test download collection',
        'description': 'Testing. Testing. 1, 2, 3.',
    },
)
class TestingDownload(ItemWithAttachment):
    item_type = 'testing_download'
    schema = {
        'type': 'object',
        'properties': {
            'attachment': {
                'type': 'object',
                'attachment': True,
                'properties': {
                    'type': {
                        'type': 'string',
                        'enum': ['image/png'],
                    }
                }
            },
            'attachment2': {
                'type': 'object',
                'attachment': True,
                'properties': {
                    'type': {
                        'type': 'string',
                        'enum': ['image/png'],
                    }
                }
            },
            'attachment3': {
                'type': 'object',
                'attachment': True,
                'properties': {
                    'type': {
                        'type': 'string',
                        'enum': ['application/json'],
                    }
                }
            },
            'attachment4': {
                'type': 'object',
                'attachment': True,
                'properties': {
                    'type': {
                        'type': 'string',
                        'enum': ['application/json'],
                    }
                }
            }
        }
    }


@collection('testing-link-sources')
class TestingLinkSource(Item):
    item_type = 'testing_link_source'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
            },
            'uuid': {
                'type': 'string',
            },
            'target': {
                'type': 'string',
                'linkTo': 'TestingLinkTarget',
            },
            'status': {
                'type': 'string',
            },
            'condition1': {
                'type': 'boolean',
            },
        },
        'required': ['target'],
        'additionalProperties': False,
    }


@collection('testing-link-targets', unique_key='testing_link_target:name')
class TestingLinkTarget(Item):
    item_type = 'testing_link_target'
    name_key = 'name'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'uniqueKey': True,
            },
            'uuid': {
                'type': 'string',
            },
            'status': {
                'type': 'string',
            },
        },
        'additionalProperties': False,
    }
    rev = {
        'reverse': ('TestingLinkSource', 'target'),
    }
    embedded = [
        'reverse',
    ]
    audit_inherit = ['*']

    @calculated_property(schema={
        "title": "Sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "TestingLinkSource.target",
        },
    })
    def reverse(self, request, reverse):
        return paths_filtered_by_status(request, reverse)


@collection(
    'testing-search-scheams',
    unique_key='testing_search_schema:name'
)
class TestingSearchSchema(Item):
    item_type = 'testing_search_schema'
    name_key = 'name'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'uniqueKey': True,
            },
            'status': {
                'type': 'string',
            },
            'uuid': {
                'title': 'UUID',
                'description': 'Unique identifier',
                'type': 'string',
                'format': 'uuid',
                'permission': 'import_items',
                'requestMethod': 'POST',
            },
            'accession': {
                'title': 'Accession',
                'description': '',
                'type': 'string',
                'format': 'accession',
                'permission': 'import_items'
            },
            'label': {
                'type': 'string',
            }
        },
        'additionalProperties': False,
        'facet_groups': [
            {
                'title': 'Test group',
                'facet_fields': [
                    'status',
                    'name'
                ]
            }
        ],
        'facets': {
            'status': {
                'title': 'Status',
                'open_on_load': True
            },
            'name': {
                'title': 'Name'
            }
        },
        'boost_values': {
            'accession': 1.0,
            'status': 1.0,
            'label': 1.0
        },
        'columns': {
            'accession': {
                'title': 'Accession'
            },
            'status': {
                'title': 'Status'
            }
        }
    }
    audit_inherit = ['*']
    matrix = {
        'x': {
            'group_by': 'label'
        },
        'y': {
            'group_by': ['status', 'name']
        }
    }
    missing_matrix = {
        'x': {
            'group_by': 'label'
        },
        'y': {
            'group_by': ['status', ('name', 'default_name')]
        }
    }
    audit = {
        'audit.ERROR.category': {
            'group_by': 'audit.ERROR.category',
            'label': 'Error'
        },
        'audit.INTERNAL_ACTION.category': {
            'group_by': 'audit.INTERNAL_ACTION.category',
            'label': 'Internal Action'},
        'audit.NOT_COMPLIANT.category': {
            'group_by': 'audit.NOT_COMPLIANT.category',
            'label': 'Not Compliant'
        },
        'audit.WARNING.category': {
            'group_by': 'audit.WARNING.category',
            'label': 'Warning'
        },
        'x': {
            'group_by': 'status', 'label': 'Status'
        }
    }


@collection(
    'testing-post-put-patch',
    acl=[
        (Allow, 'group.submitter', ['add', 'edit', 'view']),
    ],
)
class TestingPostPutPatch(Item):
    item_type = 'testing_post_put_patch'
    schema = {
        'required': ['required'],
        'type': 'object',
        'properties': {
            "schema_version": {
                "type": "string",
                "pattern": "^\\d+(\\.\\d+)*$",
                "requestMethod": [],
                "default": "1",
            },
            "uuid": {
                "title": "UUID",
                "description": "",
                "type": "string",
                "format": "uuid",
                "permission": "import_items",
                "requestMethod": "POST",
            },
            "accession": {
                "title": "Accession",
                "description": "",
                "type": "string",
                "format": "accession",
                "permission": "import_items"
            },
            'required': {
                'type': 'string',
            },
            'simple1': {
                'type': 'string',
                'default': 'simple1 default',
            },
            'simple2': {
                'type': 'string',
                'default': 'simple2 default',
            },
            'protected': {
                # This should be allowed on PUT so long as value is the same
                'type': 'string',
                'default': 'protected default',
                'permission': 'import_items',
            },
            'protected_link': {
                # This should be allowed on PUT so long as the linked uuid is
                # the same
                'type': 'string',
                'linkTo': 'TestingLinkTarget',
                'permission': 'import_items',
            },
        }
    }


@collection('testing-server-defaults')
class TestingServerDefault(Item):
    item_type = 'testing_server_default'
    schema = {
        'type': 'object',
        'properties': {
            "schema_version": {
                "type": "string",
                "pattern": "^\\d+(\\.\\d+)*$",
                "requestMethod": [],
            },
            "uuid": {
                "title": "UUID",
                "description": "",
                "type": "string",
                "format": "uuid",
                "permission": "import_items",
                "requestMethod": "POST",
            },
            'user': {
                'serverDefault': 'userid',
                'linkTo': 'User',
                'type': 'string',
            },
            'now': {
                'serverDefault': 'now',
                'format': 'date-time',
                'type': 'string',
            },
            'accession': {
                'serverDefault': 'accession',
                'accessionType': 'SS',
                'format': 'accession',
                'type': 'string',
            },
        }
    }


@collection('testing-bad-accession')
class TestingBadAccession(Item):
    item_type = 'testing_bad_accession'
    schema = {
        'type': 'object',
        'properties': {
            "schema_version": {
                "type": "string",
                "pattern": "^\\d+(\\.\\d+)*$",
            },
            "uuid": {
                "title": "UUID",
                "description": "",
                "type": "string",
                "format": "uuid",
                "permission": "import_items",
                "requestMethod": "POST",
            },
            'thing': {
                'type': "number",
                'default': 3,
            },
            'user': {
                'serverDefault': 'userid',
                'linkTo': 'User',
                'type': 'string',
            },
            'now': {
                'serverDefault': 'now',
                'format': 'date-time',
                'type': 'string',
            },
            'accession': {
                'serverDefault': 'accession',
                'accessionType': 'SS',
                'format': 'accession',
                'type': 'string',
            },
        }
    }


@collection('testing-dependencies')
class TestingDependencies(Item):
    item_type = 'testing_dependencies'
    schema = {
        'type': 'object',
        'dependencies': {
            'dep1': ['dep2'],
            'dep2': ['dep1'],
        },
        'properties': {
            'dep1': {
                'type': 'string',
            },
            'dep2': {
                'type': 'string',
                'enum': ['dep2'],
            },
        }
    }


@view_config(name='testing-render-error', request_method='GET')
def testing_render_error(request):
    return {
        '@type': ['TestingRenderError', 'Item'],
        '@id': request.path,
        'title': 'Item triggering a render error',
    }


@view_config(context=TestingPostPutPatch, name='testing-retry')
def testing_retry(context, request):
    from sqlalchemy import inspect
    from transaction.interfaces import TransientError

    model = context.model
    attempt = request.environ.get('retry.attempts')

    if attempt == 0:
        raise TransientError()

    return {
        'retry.attempts': attempt,
        'detached': inspect(model).detached,
    }


@collection('testing-custom-embed-sources')
class TestingCustomEmbedSource(Item):
    item_type = 'testing_custom_embed_source'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
            },
            'uuid': {
                'type': 'string',
            },
            'target': {
                'type': 'string',
                'linkTo': 'TestingCustomEmbedTarget',
            },
            'status': {
                'type': 'string',
            },
            'condition1': {
                'type': 'boolean',
            },
        },
        'required': ['target'],
        'additionalProperties': False,
    }


@collection('testing-custom-embed-targets', unique_key='testing_custom_embed_target:name')
class TestingCustomEmbedTarget(Item):
    item_type = 'testing_custom_embed_target'
    name_key = 'name'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'uniqueKey': True,
            },
            'uuid': {
                'type': 'string',
            },
            'status': {
                'type': 'string',
            },
        },
        'additionalProperties': False,
    }
    rev = {
        'reverse': ('TestingCustomEmbedSource', 'target'),
    }
    embedded_with_frame = [
        Path('reverse'),
        Path('filtered_reverse', include=['uuid', 'status']),
        Path('filtered_reverse1', exclude=['uuid', '@type']),
        Path('reverse_uncalculated', frame='@@object?skip_calculated=true'),
    ]
    audit_inherit = ['*']

    @calculated_property(schema={
        "title": "Sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "TestingCustomEmbedSource.target",
        },
    })
    def reverse(self, request, reverse):
        return paths_filtered_by_status(request, reverse)


    @calculated_property(schema={
        "title": "Filtered sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "TestingCustomEmbedSource.target",
        },
    })
    def filtered_reverse(self, request, reverse):
        return paths_filtered_by_status(request, reverse)

    @calculated_property(schema={
        "title": "Filtered sources1",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "TestingCustomEmbedSource.target",
        },
    })
    def filtered_reverse1(self, request, reverse):
        return paths_filtered_by_status(request, reverse)

    @calculated_property(schema={
        "title": "Uncalculated sources",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "TestingCustomEmbedSource.target",
        },
    })
    def reverse_uncalculated(self, request, reverse):
        return paths_filtered_by_status(request, reverse)


@collection(
    'testing-search-schema-special-facets',
    unique_key='testing_search_schema-special-facets:name'
)
class TestingSearchSchemaSpecialFacets(Item):
    item_type = 'testing_search_schema_special_facets'
    name_key = 'name'
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string',
                'uniqueKey': True,
            },
            'status': {
                'type': 'string',
            },
            'read_count': {
                'type': 'number',
            },
            'uuid': {
                'title': 'UUID',
                'description': 'Unique identifier',
                'type': 'string',
                'format': 'uuid',
                'permission': 'import_items',
                'requestMethod': 'POST',
            },
            'accession': {
                'title': 'Accession',
                'description': '',
                'type': 'string',
                'format': 'accession',
                'permission': 'import_items'
            },
            'label': {
                'type': 'string',
            }
        },
        'additionalProperties': False,
        'facets': {
            'status': {
                'title': 'Status',
                'type': 'exists',
            },
            'read_count': {
                'title': 'Read count range',
                'type': 'stats',
            },
            'name': {
                'title': 'Name'
            }
        },
        'boost_values': {
            'accession': 1.0,
            'status': 1.0,
            'label': 1.0
        },
        'columns': {
            'accession': {
                'title': 'Accession'
            },
            'status': {
                'title': 'Status'
            }
        }
    }


@search_config(
    name='TestConfigItem'
)
def new_item_search_config():
    return {
        'facets': {'a': 'b'}
    }



import argparse
import transaction
from pyramid.paster import get_app
from snovault.interfaces import STORAGE


def make_batch(iterable, batch_size):
    total = len(iterable)
    for i in range(0, total, batch_size):
        yield iterable[i:min(i + batch_size, total)]


def get_uuids_to_delete(infile):
    with open(infile, 'r') as f:
        uuids_to_delete = [
            line.strip()
            for line in f
            if line.strip() != ''
        ]
    print('{} uuids found.'.format(len(uuids_to_delete)))
    return uuids_to_delete


def delete_uuids(batch, storage, update):
    for rid in batch:
        try:
            print('Deleting {}'.format(rid))
            storage.delete_by_uuid(str(rid))
        except AttributeError:
            print('ERROR: UUID NOT FOUND. Skipping.')
    # Commit every batch.
    if update:
        print('Commiting')
        transaction.commit()


def run(app, infile, update, batch_size):
    if update:
        print('Transaction will commit.')
    else:
        print('Update not specified. Transaction will abort.')
    # Access delete_by_uuid function in storage.py.
    storage = app.registry[STORAGE].write
    # Get list of uuids.
    uuids_to_delete = get_uuids_to_delete(infile)
    # Chunk uuids into transaction groups.
    batch_uuids = make_batch(uuids_to_delete, batch_size)
    for i, batch in enumerate(batch_uuids):
        print('Batch {}:'.format(i))
        delete_uuids(batch, storage, update)
    if not update:
        print('Aborting')
        transaction.abort()


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--app-name', default='app', help='Pyramid app name in config.')
    parser.add_argument('--infile', required=True, help='Path to list of UUIDs to delete.')
    parser.add_argument('--update', action='store_true', help='Only commit transaction if specified.')
    parser.add_argument('--batch_size', default=1, type=int, help='Transaction commit batch_size.')
    parser.add_argument('config_uri', help='Path to config.')
    return parser.parse_args()


def main():
    args = parse_args()
    app = get_app(args.config_uri, args.app_name)
    with transaction.manager:
        run(app, args.infile, args.update, args.batch_size)


if __name__ == '__main__':
    main()



from pyramid.view import view_config

from snosearch.interfaces import AUDIT_TITLE
from snosearch.interfaces import MATRIX_TITLE
from snosearch.interfaces import REPORT_TITLE
from snosearch.interfaces import SEARCH_TITLE
from snosearch.interfaces import SUMMARY_MATRIX
from snosearch.interfaces import SUMMARY_TITLE
from snosearch.fields import AuditMatrixWithFacetsResponseField
from snosearch.fields import AllResponseField
from snosearch.fields import BasicMatrixWithFacetsResponseField
from snosearch.fields import BasicSearchResponseField
from snosearch.fields import BasicSearchWithFacetsResponseField
from snosearch.fields import BasicSearchWithoutFacetsResponseField
from snosearch.fields import BasicReportWithFacetsResponseField
from snosearch.fields import BasicReportWithoutFacetsResponseField
from snosearch.fields import CachedFacetsResponseField
from snosearch.fields import ClearFiltersResponseField
from snosearch.fields import ColumnsResponseField
from snosearch.fields import ContextResponseField
from snosearch.fields import DebugQueryResponseField
from snosearch.fields import FiltersResponseField
from snosearch.fields import IDResponseField
from snosearch.fields import MissingMatrixWithFacetsResponseField
from snosearch.fields import NotificationResponseField
from snovault.elasticsearch.searches.fields import NonSortableResponseField
from snosearch.fields import RawMatrixWithAggsResponseField
from snosearch.fields import RawSearchWithAggsResponseField
from snosearch.fields import SearchBaseResponseField
from snosearch.fields import SortResponseField
from snosearch.fields import TitleResponseField
from snosearch.fields import RawTopHitsResponseField
from snosearch.fields import TypeOnlyClearFiltersResponseField
from snosearch.fields import TypeResponseField
from snosearch.parsers import ParamsParser
from snosearch.responses import FieldedGeneratorResponse
from snosearch.responses import FieldedResponse


def includeme(config):
    config.add_route('search', '/search{slash:/?}')
    config.add_route('searchv2_raw', '/searchv2_raw{slash:/?}')
    config.add_route('searchv2_quick', '/searchv2_quick{slash:/?}')
    config.add_route('search-cached-facets', '/search-cached-facets{slash:/?}')
    config.add_route('report', '/report{slash:/?}')
    config.add_route('report-cached-facets', '/report-cached-facets{slash:/?}')
    config.add_route('matrixv2_raw', '/matrixv2_raw{slash:/?}')
    config.add_route('matrix', '/matrix{slash:/?}')
    config.add_route('missing_matrix', '/missing_matrix{slash:/?}')
    config.add_route('summary', '/summary{slash:/?}')
    config.add_route('audit', '/audit{slash:/?}')
    config.add_route('top_hits_raw', '/top_hits_raw{slash:/?}')
    config.scan(__name__)


DEFAULT_ITEM_TYPES = [
    'Lab',
    'Snowset',
    'Snowball',
    'Snowfort',
    'Snowflake',
]


@view_config(route_name='search', request_method='GET', permission='search')
def search(context, request):
    # Note the order of rendering matters for some fields, e.g. AllResponseField and
    # NotificationResponseField depend on results from BasicSearchWithFacetsResponseField.
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=SEARCH_TITLE
            ),
            TypeResponseField(
                at_type=[SEARCH_TITLE]
            ),
            IDResponseField(),
            ContextResponseField(),
            BasicSearchWithFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            ),
            AllResponseField(),
            NotificationResponseField(),
            FiltersResponseField(),
            ClearFiltersResponseField(),
            ColumnsResponseField(),
            SortResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='searchv2_raw', request_method='GET', permission='search')
def searchv2_raw(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            RawSearchWithAggsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            )
        ]
    )
    return fr.render()


@view_config(route_name='searchv2_quick', request_method='GET', permission='search')
def searchv2_quick(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            BasicSearchResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            )
        ]
    )
    return fr.render()


@view_config(route_name='search-cached-facets', request_method='GET', permission='search')
def search_cached_facets(context, request):
    # Note the order of rendering matters for some fields, e.g. AllResponseField and
    # NotificationResponseField depend on results from BasicSearchWithFacetsResponseField.
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=SEARCH_TITLE
            ),
            TypeResponseField(
                at_type=[SEARCH_TITLE]
            ),
            IDResponseField(),
            ContextResponseField(),
            CachedFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            ),
            BasicSearchWithoutFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            ),
            AllResponseField(),
            NotificationResponseField(),
            FiltersResponseField(),
            ClearFiltersResponseField(),
            ColumnsResponseField(),
            SortResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


def search_generator(request):
    '''
    For internal use (no view). Like search_quick but returns raw generator
    of search hits in @graph field.
    '''
    fgr = FieldedGeneratorResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            BasicSearchResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            )
        ]
    )
    return fgr.render()


@view_config(route_name='report', request_method='GET', permission='search')
def report(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=REPORT_TITLE
            ),
            TypeResponseField(
                at_type=[REPORT_TITLE]
            ),
            IDResponseField(),
            ContextResponseField(),
            BasicReportWithFacetsResponseField(),
            AllResponseField(),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            ColumnsResponseField(),
            NonSortableResponseField(),
            SortResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='report-cached-facets', request_method='GET', permission='search')
def report_cached_facets(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=REPORT_TITLE
            ),
            TypeResponseField(
                at_type=[REPORT_TITLE]
            ),
            IDResponseField(),
            ContextResponseField(),
            CachedFacetsResponseField(),
            BasicReportWithoutFacetsResponseField(),
            AllResponseField(),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            ColumnsResponseField(),
            NonSortableResponseField(),
            SortResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='matrixv2_raw', request_method='GET', permission='search')
def matrixv2_raw(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            RawMatrixWithAggsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            )
        ]
    )
    return fr.render()


@view_config(route_name='matrix', request_method='GET', permission='search')
def matrix(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=MATRIX_TITLE
            ),
            TypeResponseField(
                at_type=[MATRIX_TITLE]
            ),
            IDResponseField(),
            SearchBaseResponseField(),
            ContextResponseField(),
            BasicMatrixWithFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            ),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='missing_matrix', request_method='GET', permission='search')
def missing_matrix(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=MATRIX_TITLE
            ),
            TypeResponseField(
                at_type=[MATRIX_TITLE]
            ),
            IDResponseField(),
            SearchBaseResponseField(),
            ContextResponseField(),
            MissingMatrixWithFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES,
                matrix_definition_name='missing_matrix'
            ),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='summary', request_method='GET', permission='search')
def summary(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=SUMMARY_TITLE
            ),
            TypeResponseField(
                at_type=[SUMMARY_TITLE]
            ),
            IDResponseField(),
            SearchBaseResponseField(),
            ContextResponseField(),
            BasicMatrixWithFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES,
                matrix_definition_name=SUMMARY_MATRIX
            ),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='audit', request_method='GET', permission='search')
def audit(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            TitleResponseField(
                title=AUDIT_TITLE
            ),
            TypeResponseField(
                at_type=[AUDIT_TITLE]
            ),
            IDResponseField(),
            SearchBaseResponseField(),
            ContextResponseField(),
            AuditMatrixWithFacetsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            ),
            NotificationResponseField(),
            FiltersResponseField(),
            TypeOnlyClearFiltersResponseField(),
            DebugQueryResponseField()
        ]
    )
    return fr.render()


@view_config(route_name='top_hits_raw', request_method='GET', permission='search')
def top_hits_raw(context, request):
    fr = FieldedResponse(
        _meta={
            'params_parser': ParamsParser(request)
        },
        response_fields=[
            RawTopHitsResponseField(
                default_item_types=DEFAULT_ITEM_TYPES
            )
        ]
    )
    return fr.render()



import pytest
import os

from time import sleep
from subprocess import TimeoutExpired

from pyramid.paster import get_appsettings


def pytest_configure():
    import logging
    logging.basicConfig()
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)

    class Shorten(logging.Filter):
        max_len = 500

        def filter(self, record):
            if record.msg == '%r':
                record.msg = record.msg % record.args
                record.args = ()
            if len(record.msg) > self.max_len:
                record.msg = record.msg[:self.max_len] + '...'
            return True

    logging.getLogger('sqlalchemy.engine.base.Engine').addFilter(Shorten())


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def engine_url(request):
    # Ideally this would use a different database on the same postgres server
    from urllib.parse import quote
    from .postgresql_fixture import initdb, server_process
    tmpdir = request.config._tmpdirhandler.mktemp('postgresql-engine', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = server_process(tmpdir)

    yield 'postgresql://postgres@:5432/postgres?host=%s' % quote(tmpdir)

    if process.poll() is None:
        process.terminate()
        process.wait()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def postgresql_server(request):
    from urllib.parse import quote
    from .postgresql_fixture import initdb, server_process
    tmpdir = request.config._tmpdirhandler.mktemp('postgresql', numbered=True)
    tmpdir = str(tmpdir)
    initdb(tmpdir)
    process = server_process(tmpdir)

    yield 'postgresql://postgres@:5432/postgres?host=%s' % quote(tmpdir)

    if process.poll() is None:
        process.terminate()
        process.wait()


@pytest.fixture(scope='session')
def elasticsearch_host_port():
    from webtest.http import get_free_port
    return get_free_port()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def elasticsearch_server(request, elasticsearch_host_port):
    from .elasticsearch_fixture import server_process
    host, port = elasticsearch_host_port
    tmpdir = request.config._tmpdirhandler.mktemp('elasticsearch', numbered=True)
    tmpdir = str(tmpdir)
    process = server_process(str(tmpdir), host=host, port=9201, echo=False)
    print('PORT CHANGED')
    yield 'http://%s:%d' % (host, 9201)

    if 'process' in locals() and process.poll() is None:
        process.terminate()
        try:
            process.wait(timeout=10)
        except TimeoutExpired:
            process.kill()


@pytest.mark.fixture_cost(10)
@pytest.yield_fixture(scope='session')
def redis_server(request):
    from .redis_storage_fixture import initdb, server_process
    datadir = str(request.config._tmpdirhandler.mktemp('redisdatatest', numbered=True))
    appsettings = get_appsettings('development.ini', name='app')
    # Required settings in config
    local_storage_host = appsettings['local_storage_host']
    local_storage_port = appsettings['local_storage_port']
    local_storage_redis_index = appsettings['local_storage_redis_index']
    local_storage_timeout = appsettings['local_storage_timeout']
    # Build fixture
    redis_config_path = initdb(datadir, local_storage_port, echo=True)
    process = server_process(redis_config_path, local_storage_port, local_storage_redis_index, echo=True)
    # Sleep for short time to allow redis db to initialize
    sleep(0.25)
    yield f"Redis testing: redis-cli -p {local_storage_port} -n {local_storage_redis_index})"
    if 'process' in locals() and process.poll() is None:
        process.terminate()
        try:
            process.wait(timeout=10)
        except TimeoutExpired:
            process.kill()


# http://docs.sqlalchemy.org/en/rel_0_8/orm/session.html#joining-a-session-into-an-external-transaction
# By binding the SQLAlchemy Session to an external transaction multiple testapp
# requests can be rolled back at the end of the test.

@pytest.yield_fixture(scope='session')
def conn(engine_url):
    from snovault.app import configure_engine
    from snovault.storage import Base

    engine_settings = {
        'sqlalchemy.url': engine_url,
    }

    engine = configure_engine(engine_settings)
    conn = engine.connect()
    tx = conn.begin()
    try:
        Base.metadata.create_all(bind=conn)
        yield conn
    finally:
        tx.rollback()
        conn.close()
        engine.dispose()


@pytest.fixture(scope='session')
def _DBSession(conn):
    import snovault.storage
    import zope.sqlalchemy
    from sqlalchemy import orm
    # ``server`` thread must be in same scope
    DBSession = orm.scoped_session(orm.sessionmaker(bind=conn), scopefunc=lambda: 0)
    zope.sqlalchemy.register(DBSession)
    snovault.storage.register(DBSession)
    return DBSession


@pytest.fixture(scope='session')
def DBSession(_DBSession, zsa_savepoints, check_constraints):
    return _DBSession


@pytest.yield_fixture
def external_tx(request, conn):
    # print('BEGIN external_tx')
    tx = conn.begin_nested()
    yield tx
    tx.rollback()
    # # The database should be empty unless a data fixture was loaded
    # from snovault.storage import Base
    # for table in Base.metadata.sorted_tables:
    #     assert conn.execute(table.count()).scalar() == 0


@pytest.fixture
def transaction(request, external_tx, zsa_savepoints, check_constraints):
    import transaction
    transaction.begin()
    request.addfinalizer(transaction.abort)
    return transaction


@pytest.yield_fixture(scope='session')
def zsa_savepoints(conn):
    """ Place a savepoint at the start of the zope transaction

    This means failed requests rollback to the db state when they began rather
    than that at the start of the test.
    """
    from transaction.interfaces import ISynchronizer
    from zope.interface import implementer

    @implementer(ISynchronizer)
    class Savepoints(object):
        def __init__(self, conn):
            self.conn = conn
            self.sp = None
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            # txn be aborted a second time in manager.begin()
            if self.sp is None:
                return
            if self.state == 'commit':
                self.state = 'completion'
                self.sp.commit()
            else:
                self.state = 'abort'
                self.sp.rollback()
            self.sp = None
            self.state = 'done'

        def newTransaction(self, transaction):
            self.state = 'new'
            self.sp = self.conn.begin_nested()
            self.state = 'begun'
            transaction.addBeforeCommitHook(self._registerCommit)

        def _registerCommit(self):
            self.state = 'commit'

    zsa_savepoints = Savepoints(conn)

    import transaction
    transaction.manager.registerSynch(zsa_savepoints)

    yield zsa_savepoints
    transaction.manager.unregisterSynch(zsa_savepoints)


@pytest.fixture
def session(transaction, DBSession):
    """ Returns a setup session

    Depends on transaction as storage relies on some interaction there.
    """
    return DBSession()


@pytest.yield_fixture(scope='session')
def check_constraints(conn, _DBSession):
    '''Check deffered constraints on zope transaction commit.

    Deferred foreign key constraints are only checked at the outer transaction
    boundary, not at a savepoint. With the Pyramid transaction bound to a
    subtransaction check them manually.
    '''
    from transaction.interfaces import ISynchronizer
    from zope.interface import implementer

    @implementer(ISynchronizer)
    class CheckConstraints(object):
        def __init__(self, conn):
            self.conn = conn
            self.state = None

        def beforeCompletion(self, transaction):
            pass

        def afterCompletion(self, transaction):
            pass

        def newTransaction(self, transaction):

            @transaction.addBeforeCommitHook
            def set_constraints():
                self.state = 'checking'
                session = _DBSession()
                session.flush()
                sp = self.conn.begin_nested()
                try:
                    self.conn.execute('SET CONSTRAINTS ALL IMMEDIATE')
                except:
                    sp.rollback()
                    raise
                else:
                    self.conn.execute('SET CONSTRAINTS ALL DEFERRED')
                finally:
                    sp.commit()
                    self.state = None

    check_constraints = CheckConstraints(conn)

    import transaction
    transaction.manager.registerSynch(check_constraints)

    yield check_constraints

    transaction.manager.unregisterSynch(check_constraints)


@pytest.yield_fixture
def execute_counter(conn, zsa_savepoints, check_constraints):
    """ Count calls to execute
    """
    from contextlib import contextmanager
    from sqlalchemy import event

    class Counter(object):
        def __init__(self):
            self.reset()
            self.conn = conn

        def reset(self):
            self.count = 0

        @contextmanager
        def expect(self, count):
            start = self.count
            yield
            difference = self.count - start
            assert difference == count

    counter = Counter()

    @event.listens_for(conn, 'after_cursor_execute')
    def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
        # Ignore the testing savepoints
        if zsa_savepoints.state != 'begun' or check_constraints.state == 'checking':
            return
        counter.count += 1

    yield counter

    event.remove(conn, 'after_cursor_execute', after_cursor_execute)


@pytest.yield_fixture
def no_deps(conn, DBSession):
    from sqlalchemy import event

    session = DBSession()

    @event.listens_for(session, 'after_flush')
    def check_dependencies(session, flush_context):
        assert not flush_context.cycles

    @event.listens_for(conn, "before_execute", retval=True)
    def before_execute(conn, clauseelement, multiparams, params):
        return clauseelement, multiparams, params

    yield

    event.remove(session, 'before_flush', check_dependencies)


@pytest.fixture(scope='session')
def wsgi_server_host_port(request):
    wsgi_args = dict(request.config.option.wsgi_args or ())
    if ('port_range.min' in wsgi_args and 'port_range.max' in wsgi_args):
        import socket
        import os
        # return available port in specified range if min and max are defined
        port_temp, port_max = int(wsgi_args['port_range.min']), int(wsgi_args['port_range.max'])
        port_assigned = False
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while not port_assigned and port_temp <= port_max:
            try:
                s.bind(('', port_temp))
                port_assigned = True
            except OSError:
                port_temp += 1
        if not port_assigned:
            # port failed to be assigned, so raise an error
            raise
        ip, port = s.getsockname()
        s.close()
        ip = os.environ.get('WEBTEST_SERVER_BIND', '127.0.0.1')
        return ip, port
    else:
        # otherwise get any free port
        from webtest.http import get_free_port
        return get_free_port()


@pytest.fixture(scope='session')
def wsgi_server_app(app):
    return app


@pytest.mark.fixture_cost(100)
@pytest.yield_fixture(scope='session')
def wsgi_server(request, wsgi_server_app, wsgi_server_host_port):
    from webtest.http import StopableWSGIServer
    host, port = wsgi_server_host_port

    server = StopableWSGIServer.create(
        wsgi_server_app,
        host=host,
        port=port,
        threads=1,
        channel_timeout=60,
        cleanup_interval=10,
        expose_tracebacks=True,
        clear_untrusted_proxy_headers=False,
    )
    assert server.wait()
    print("wsgi server port {}".format(port))

    yield 'http://%s:%s' % wsgi_server_host_port

    server.shutdown()



from past.builtins import basestring
from pyramid.settings import asbool
from pyramid.traversal import (
    find_resource,
)
from pyramid.view import view_config
from uuid import (
    UUID,
    uuid4,
)
from .etag import if_match_tid
from .interfaces import (
    COLLECTIONS,
    CONNECTION,
    Created,
    BeforeModified,
    AfterModified,
)
from .resources import (
    Collection,
    Item,
)
from .validation import ValidationFailure
from .validators import (
    no_validate_item_content_patch,
    no_validate_item_content_post,
    no_validate_item_content_put,
    validate_item_content_patch,
    validate_item_content_post,
    validate_item_content_put,
)


def includeme(config):
    config.scan(__name__)


def split_child_props(type_info, properties):
    propname_children = {}
    item_properties = properties.copy()
    if type_info.schema_rev_links:
        for key, spec in type_info.schema_rev_links.items():
            if key in item_properties:
                propname_children[key] = item_properties.pop(key)
    return item_properties, propname_children


def ensure_child_link(props, type_info, link_attr, uuid):
    if type_info.schema['properties'][link_attr]['type'] == 'array':
        if link_attr not in props:
            props[link_attr] = uuid
        elif uuid not in props:
            props[link_attr].append(uuid)
    else:
        props[link_attr] = uuid


def update_children(context, request, propname_children):
    registry = request.registry
    conn = registry[CONNECTION]
    collections = registry[COLLECTIONS]
    schema_rev_links = context.type_info.schema_rev_links
    uuid = str(context.uuid)

    for propname, children in propname_children.items():
        link_type, link_attr = schema_rev_links[propname]
        found = set()

        # Add or update children included in properties
        for i, child_props in enumerate(children):
            if isinstance(child_props, basestring):  # IRI of (existing) child
                child = find_resource(collections[link_type], child_props)
            else:
                child_props = child_props.copy()
                child_type = child_props.pop('@type', None)
                if 'uuid' in child_props:  # update existing child
                    child_id = child_props.pop('uuid')
                    child = conn.get_by_uuid(child_id)
                    if not request.has_permission('edit', child):
                        msg = u'edit forbidden to %s' % request.resource_path(child)
                        raise ValidationFailure('body', [propname, i], msg)
                    ensure_child_link(
                        child_props, child.type_info, link_attr, uuid)
                    try:
                        update_item(child, request, child_props)
                    except ValidationFailure as e:
                        e.location = [propname, i] + e.location
                        raise
                else:  # add new child
                    child_type = child_type[0]
                    child_collection = collections[child_type]
                    ensure_child_link(
                        child_props, child_collection.type_info,
                        link_attr, uuid)
                    if not request.has_permission('add', child_collection):
                        msg = u'add forbidden to %s' % request.resource_path(child_collection)
                        raise ValidationFailure('body', [propname, i], msg)
                    child = create_item(child_collection.type_info, request, child_props)
            found.add(child.uuid)

        # Remove existing children that are not in properties
        for link_uuid in context.get_rev_links(propname):
            if link_uuid in found:
                continue
            child = conn.get_by_uuid(link_uuid)
            if not request.has_permission('visible_for_edit', child):
                continue
            if not request.has_permission('edit', child):
                msg = u'edit forbidden to %s' % request.resource_path(child)
                raise ValidationFailure('body', [propname, i], msg)
            try:
                delete_item(child, request)
            except ValidationFailure as e:
                e.location = [propname, i] + e.location
                raise


def create_item(type_info, request, properties, sheets=None):
    registry = request.registry
    item_properties, propname_children = split_child_props(type_info, properties)

    if 'uuid' in item_properties:
        uuid = UUID(item_properties.pop('uuid'))
    else:
        uuid = uuid4()

    item = type_info.factory.create(registry, uuid, item_properties, sheets)
    registry.notify(Created(item, request))

    if propname_children:
        update_children(item, request, propname_children)
    return item


def update_item(context, request, properties, sheets=None):
    registry = request.registry
    item_properties, propname_children = split_child_props(context.type_info, properties)

    registry.notify(BeforeModified(context, request))
    context.update(item_properties, sheets)
    registry.notify(AfterModified(context, request))

    if propname_children:
        update_children(context, request, propname_children)


def delete_item(context, request):
    properties = context.properties.copy()
    properties['status'] = 'deleted'
    update_item(context, request, properties)


@view_config(context=Collection, permission='add', request_method='POST',
             validators=[validate_item_content_post])
@view_config(context=Collection, permission='add_unvalidated', request_method='POST',
             validators=[no_validate_item_content_post],
             request_param=['validate=false'])
def collection_add(context, request, render=None):
    if render is None:
        render = request.params.get('render', True)

    item = create_item(context.type_info, request, request.validated)

    if render == 'uuid':
        item_uri = '/%s/' % item.uuid
    else:
        item_uri = request.resource_path(item)
    if asbool(render) is True:
        rendered = request.embed(item_uri, '@@object', as_user=True)
    else:
        rendered = item_uri
    request.response.status = 201
    request.response.location = item_uri
    result = {
        'status': 'success',
        '@type': ['result'],
        '@graph': [rendered],
    }
    return result


@view_config(context=Item, permission='edit', request_method='PUT',
             validators=[validate_item_content_put], decorator=if_match_tid)
@view_config(context=Item, permission='edit', request_method='PATCH',
             validators=[validate_item_content_patch], decorator=if_match_tid)
@view_config(context=Item, permission='edit_unvalidated', request_method='PUT',
             validators=[no_validate_item_content_put],
             request_param=['validate=false'], decorator=if_match_tid)
@view_config(context=Item, permission='edit_unvalidated', request_method='PATCH',
             validators=[no_validate_item_content_patch],
             request_param=['validate=false'], decorator=if_match_tid)
def item_edit(context, request, render=None):
    """ This handles both PUT and PATCH, difference is the validator

    PUT - replaces the current properties with the new body
    PATCH - updates the current properties with those supplied.
    """
    if render is None:
        render = request.params.get('render', True)

    # This *sets* the property sheet
    update_item(context, request, request.validated)

    if render == 'uuid':
        item_uri = '/%s' % context.uuid
    else:
        item_uri = request.resource_path(context)
    if asbool(render) is True:
        rendered = request.embed(item_uri, '@@object', as_user=True)
    else:
        rendered = item_uri
    request.response.status = 200
    result = {
        'status': 'success',
        '@type': ['result'],
        '@graph': [rendered],
    }
    return result




# import pytest
# from pytest_bdd import scenarios

# pytestmark = [
#     pytest.mark.bdd,
#     pytest.mark.usefixtures('workbook', 'submitter_user'),
# ]


# scenarios(
#     'user.feature',
#     strict_gherkin=False
# )



from snovault import (
    abstract_collection,
    calculated_property,
    collection,
    load_schema,
)
from .base import (
    Item,
    paths_filtered_by_status,
)


import datetime


def item_is_revoked(request, path):
    return request.embed(path, '@@object').get('status') == 'revoked'


@abstract_collection(
    name='snowsets',
    unique_key='accession',
    properties={
        'title': "Snowsets",
        'description': 'Abstract class describing different collections of snowflakes.',
    })
class Snowset(Item):
    base_types = ['Snowset'] + Item.base_types
    embedded = [
        'snowflakes',
        'snowflakes.submitted_by',
        'snowflakes.lab',
        'submitted_by',
        'lab',
        'award',
    ]
    audit_inherit = [
        'snowflakes',
        'submitted_by',
        'lab',
        'award',
    ]
    name_key = 'accession'
    rev = {
        'snowflakes': ('Snowflake', 'snowset'),
    }

    @calculated_property(condition='date_released', schema={
        "title": "Month released",
        "type": "string",
    })
    def month_released(self, date_released):
        return datetime.datetime.strptime(date_released, '%Y-%m-%d').strftime('%B, %Y')

    @calculated_property(schema={
        "title": "snowflakes",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "Snowflake.snowset",
        },
    })
    def snowflakes(self, request, snowflakes):
        return paths_filtered_by_status(request, snowflakes)


@collection(
    name='snowballs',
    unique_key='accession',
    properties={
        'title': "Snowball style snowset",
        'description': 'A set of snowflakes packed into a snowball.',
    })
class Snowball(Snowset):
    item_type = 'snowball'
    schema = load_schema('snowflakes:schemas/snowball.json')

    @calculated_property(
        schema={
        "title": "test_calculated",
        "type": "string",
        },
        define=True
    )
    def test_calculated(self):
        return 'test_calculated_value'

    @calculated_property(
        schema={
        "title": "another_test_calculated",
        "type": "string",
    })
    def another_test_calculated(self):
        return 'another_test_calculated_value'
    
    @calculated_property(
        schema={
        "title": "conditional_test_calculated",
        "type": "string",
        },
        condition='test_calculated'
    )
    def conditional_test_calculated(self):
        return 'conditional_test_calculated_value'
    
    matrix = {
        'x': {
            'group_by': 'snowflakes.type'
        },
        'y': {
            'group_by': ['award.rfa', 'lab.title']
        }
    }
    
    missing_matrix = {
        'x': {
            'group_by': 'snowflakes.type'
        },
        'y': {
            'group_by': ['award.rfa', ('lab.not_a_real_value', 'some_lab')]
        }
    }
    
    summary_matrix = {
        'x': {
            'group_by': 'status'
        },
        'y': {
            'group_by': ['snowflakes.type']
        }
    }

    audit = {
        'x': {
            'group_by': 'snowflakes.type',
            'label': 'Type',
        },
        'audit.ERROR.category': {
            'group_by': 'audit.ERROR.category',
            'label': 'Error',
        },
        'audit.NOT_COMPLIANT.category': {
            'group_by': 'audit.NOT_COMPLIANT.category',
            'label': 'Not Compliant',
        },
        'audit.WARNING.category': {
            'group_by': 'audit.WARNING.category',
            'label': 'Warning',
        },
        'audit.INTERNAL_ACTION.category': {
            'group_by': 'audit.INTERNAL_ACTION.category',
            'label': 'Internal Action',
        },
    }


@collection(
    name='snowforts',
    unique_key='accession',
    properties={
        'title': "Snowfort style snowset",
        'description': 'A set of snowflakes packed into a snowfort.',
    })
class Snowfort(Snowset):
    item_type = 'snowfort'
    schema = load_schema('snowflakes:schemas/snowfort.json')


@collection(
    name='snowflakes',
    unique_key='accession',
    properties={
        'title': 'Snowflakes',
        'description': 'Listing of Snowflakes',
    })
class Snowflake(Item):
    item_type = 'snowflake'
    schema = load_schema('snowflakes:schemas/snowflake.json')
    name_key = 'accession'

    embedded = [
        'lab',
        'submitted_by',

    ]
    audit_inherit = [
        'lab',
        'submitted_by',

    ]



from snovault import (
    AuditFailure,
    audit_checker,
)
from snovault import (
    UPGRADER,
)
from snovault.schema_utils import validate
from snovault.util import simple_path_ids


@audit_checker('Item', frame='object')
def audit_item_schema(value, system):
    context = system['context']
    registry = system['registry']
    if not context.schema:
        return

    properties = context.properties.copy()
    current_version = properties.get('schema_version', '')
    target_version = context.type_info.schema_version
    if target_version is not None and current_version != target_version:
        upgrader = registry[UPGRADER]
        try:
            properties = upgrader.upgrade(
                context.type_info.name, properties, current_version, target_version,
                finalize=False, context=context, registry=registry)
        except RuntimeError:
            raise
        except Exception as e:
            detail = '%r upgrading from %r to %r' % (e, current_version, target_version)
            yield AuditFailure('upgrade failure', detail, level='INTERNAL_ACTION')
            return

        properties['schema_version'] = target_version

    properties['uuid'] = str(context.uuid)
    validated, errors = validate(context.schema, properties, properties)
    for error in errors:
        category = 'validation error'
        path = list(error.path)
        if path:
            category += ': ' + '/'.join(str(elem) for elem in path)
        detail = 'Object {} has schema error {}'.format(value['@id'], error.message)
        yield AuditFailure(category, detail, level='INTERNAL_ACTION')


STATUS_LEVEL = {
    # public statuses
    'released': 100,
    'current': 100,

    # 'discouraged for use' public statuses
    'archived': 40,
    'revoked': 30,

    # private statuses (visible for consortium members only)
    'in progress': 50,
    'proposed': 50,
    'started': 50,
    'submitted': 50,

    # invisible statuses (visible for admins only)
    'deleted': 0,
    'replaced': 0,
    'disabled': 0
    }


@audit_checker('Item', frame='object')
def audit_item_relations_status(value, system):
    if 'status' not in value:
        return

    level = STATUS_LEVEL.get(value['status'], 50)

    context = system['context']
    request = system['request']

    for schema_path in context.type_info.schema_links:
        if schema_path in ['supersedes']:
            for path in simple_path_ids(value, schema_path):
                linked_value = request.embed(path + '@@object')
                if 'status' not in linked_value:
                    continue
                else:
                    linked_level = STATUS_LEVEL.get(
                        linked_value['status'], 50)
                    detail = \
                        '{} with status \'{}\' supersedes {} with status \'{}\''.format(
                            value['@id'],
                            value['status'],
                            linked_value['@id'],
                            linked_value['status']
                            )
                    if level == 100 and linked_level in [0, 50, 100]:
                        yield AuditFailure(
                            'mismatched status',
                            detail,
                            level='INTERNAL_ACTION')
                    elif level == 50 and linked_level in [0, 50]:
                        yield AuditFailure(
                            'mismatched status',
                            detail,
                            level='INTERNAL_ACTION')
                    elif level in [30, 40] and linked_level in [0, 50, 100]:
                        yield AuditFailure(
                            'mismatched status',
                            detail,
                            level='INTERNAL_ACTION')

        elif schema_path in ['derived_from',
                             'controlled_by',
                             'possible_controls']:
            message = 'has a possible control'
            if schema_path == 'derived_from':
                message = 'is derived from'
            elif schema_path == 'controlled_by':
                message = 'is controlled by'
            for path in simple_path_ids(value, schema_path):
                linked_value = request.embed(path + '@@object')
                if 'status' not in linked_value:
                    continue
                else:
                    linked_level = STATUS_LEVEL.get(
                        linked_value['status'], 50)
                    if level > linked_level:
                        detail = \
                            '{} with status \'{}\' {} {} with status \'{}\''.format(
                                value['@id'],
                                value['status'],
                                message,
                                linked_value['@id'],
                                linked_value['status']
                                )
                        yield AuditFailure(
                            'mismatched status',
                            detail,
                            level='INTERNAL_ACTION')


@audit_checker('Item', frame='object')
def audit_item_status(value, system):
    if 'status' not in value:
        return

    level = STATUS_LEVEL.get(value['status'], 50)

    if level == 0:
        return

    if value['status'] in ['revoked', 'archived']:
        level += 50

    context = system['context']
    request = system['request']
    linked = set()

    for schema_path in context.type_info.schema_links:
        if schema_path in ['supersedes',
                           'step_run',
                           'derived_from',
                           'controlled_by',
                           'possible_controls']:
            continue
        else:
            linked.update(simple_path_ids(value, schema_path))

    for path in linked:
        linked_value = request.embed(path + '@@object')
        if 'status' not in linked_value:
            continue
        if linked_value['status'] == 'disabled':
            continue
        if (  # Special case: A revoked file can have a deleted replicate ticket #2938
            'File' in value['@type'] and
            value['status'] == 'revoked' and
            'Replicate' in linked_value['@type'] and
            linked_value['status'] == 'deleted'
        ):
            continue
        linked_level = STATUS_LEVEL.get(linked_value['status'], 50)
        if linked_value['status'] in ['revoked', 'archived']:
            linked_level += 50
        if linked_level == 0:
            detail = '{} {} has {} subobject {}'.format(
                value['status'], value['@id'], linked_value['status'], linked_value['@id'])
            yield AuditFailure('mismatched status', detail, level='INTERNAL_ACTION')
        elif linked_level < level:
            detail = '{} {} has {} subobject {}'.format(
                value['status'], value['@id'], linked_value['status'], linked_value['@id'])
            yield AuditFailure('mismatched status', detail, level='INTERNAL_ACTION')



from snovault.auditor import (
    audit_checker,
    AuditFailure,
)


def includeme(config):
    config.scan('.testing_views')
    config.scan(__name__)


def has_condition1(value, system):
    return value.get('condition1')


@audit_checker('testing_link_source', condition=has_condition1)
def checker1(value, system):
    if not value.get('checker1'):
        return AuditFailure('testchecker', 'Missing checker1')


@audit_checker('testing_link_target')
def testing_link_target_status(value, system):
    if value.get('status') == 'CHECK':
        if not len(value['reverse']):
            return AuditFailure('status', 'Missing reverse items')



from pkg_resources import resource_filename
from pyramid.events import (
    BeforeRender,
    subscriber,
)
from pyramid.httpexceptions import (
    HTTPMovedPermanently,
    HTTPPreconditionFailed,
    HTTPUnauthorized,
    HTTPUnsupportedMediaType,
)
from pyramid.security import forget
from pyramid.settings import asbool
from pyramid.threadlocal import (
    manager,
)
from pyramid.traversal import (
    split_path_info,
    _join_path_tuple,
)
from pyramid.tweens import EXCVIEW

from snovault.validation import CSRFTokenError
from subprocess_middleware.tween import SubprocessTween
import logging
import os
import psutil
import time


log = logging.getLogger(__name__)

# TODO Move to SnoVault?


def includeme(config):
    config.add_tween(
        '.renderers.fix_request_method_tween_factory',
        under='snovault.stats.stats_tween_factory')
    config.add_tween(
        '.renderers.normalize_cookie_tween_factory',
        under='.renderers.fix_request_method_tween_factory')

    renderer_tween = (
        '.renderers.debug_page_or_json'
        if config.registry.settings['pyramid.reload_templates']
        else '.renderers.page_or_json'
    )

    config.add_tween(
        renderer_tween,
        under='.renderers.normalize_cookie_tween_factory')

    config.add_tween(
        '.renderers.set_x_request_url_tween_factory',
        under=renderer_tween,
    )

    config.add_tween('.renderers.security_tween_factory', under=EXCVIEW)
    config.scan(__name__)


def fix_request_method_tween_factory(handler, registry):
    """ Fix Request method changed by mod_wsgi.

    See: https://github.com/GrahamDumpleton/mod_wsgi/issues/2

    Apache config:
        SetEnvIf Request_Method HEAD X_REQUEST_METHOD=HEAD
    """

    def fix_request_method_tween(request):
        environ = request.environ
        if 'X_REQUEST_METHOD' in environ:
            environ['REQUEST_METHOD'] = environ['X_REQUEST_METHOD']
        return handler(request)

    return fix_request_method_tween


def security_tween_factory(handler, registry):

    def security_tween(request):
        login = None
        expected_user = request.headers.get('X-If-Match-User')
        if expected_user is not None:
            login = request.authenticated_userid
            if login != 'mailto.' + expected_user:
                detail = 'X-If-Match-User does not match'
                raise HTTPPreconditionFailed(detail)

        # wget may only send credentials following a challenge response.
        auth_challenge = asbool(request.headers.get('X-Auth-Challenge', False))
        if auth_challenge or request.authorization is not None:
            login = request.authenticated_userid
            if login is None:
                raise HTTPUnauthorized(headerlist=forget(request))

        if request.method in ('GET', 'HEAD'):
            return handler(request)

        if request.content_type != 'application/json':
            detail = "%s is not 'application/json'" % request.content_type
            raise HTTPUnsupportedMediaType(detail)

        token = request.headers.get('X-CSRF-Token')
        if token is not None:
            # Avoid dirtying the session and adding a Set-Cookie header
            # XXX Should consider if this is a good idea or not and timeouts
            if token == dict.get(request.session, '_csrft_', None):
                return handler(request)
            raise CSRFTokenError('Incorrect CSRF token')

        # NOTE: cutting out CSRF protection here ... why protect against CSRF if you provide an
        # unathenticated endpoint that will delivery the CSRF token? I'm looking at you /session.
        # this should be revisted, either embed the csrf token in the index.html as part of the
        # rendering subprocess somehow, or return it from the login view and let the client store it
        # but of course that sounds a lot like JWT...
        return handler(request)

        if login is None:
            login = request.authenticated_userid
        if login is not None:
            namespace, userid = login.split('.', 1)
            if namespace not in ('mailto', 'persona'):
                return handler(request)
        raise CSRFTokenError('Missing CSRF token')

    return security_tween


def normalize_cookie_tween_factory(handler, registry):
    from webob.cookies import Cookie

    ignore = {
        '/favicon.ico',
    }

    def normalize_cookie_tween(request):
        if request.path in ignore or request.path.startswith('/static/'):
            return handler(request)

        session = request.session
        if session or session._cookie_name not in request.cookies:
            return handler(request)

        response = handler(request)
        existing = response.headers.getall('Set-Cookie')
        if existing:
            cookies = Cookie()
            for header in existing:
                cookies.load(header)
            if session._cookie_name in cookies:
                return response

        response.delete_cookie(
            session._cookie_name,
            path=session._cookie_path,
            domain=session._cookie_domain,
        )

        return response

    return normalize_cookie_tween


def set_x_request_url_tween_factory(handler, registry):

    def set_x_request_url_tween(request):
        response = handler(request)
        response.headers['X-Request-URL'] = request.url
        return response

    return set_x_request_url_tween


@subscriber(BeforeRender)
def canonical_redirect(event):
    request = event['request']

    # Ignore subrequests
    if len(manager.stack) > 1:
        return

    if request.method not in ('GET', 'HEAD'):
        return
    if request.response.status_int != 200:
        return
    if not request.environ.get('snowflakes.canonical_redirect', True):
        return
    if request.path_info == '/':
        return

    if not isinstance(event.rendering_val, dict):
        return

    canonical = event.rendering_val.get('@id', None)
    if canonical is None:
        return
    canonical_path, _, canonical_qs = canonical.partition('?')

    request_path = _join_path_tuple(('',) + split_path_info(request.path_info))
    if (request_path == canonical_path.rstrip('/') and
            request.path_info.endswith('/') == canonical_path.endswith('/') and
            (canonical_qs in ('', request.query_string))):
        return

    if '/@@' in request.path_info:
        return

    qs = canonical_qs or request.query_string
    location = canonical_path + ('?' if qs else '') + qs
    raise HTTPMovedPermanently(location=location)


def should_transform(request, response):
    if request.method not in ('GET', 'HEAD'):
        return False

    if response.content_type != 'application/json':
        return False

    format = request.params.get('format')
    if format is None:
        original_vary = response.vary or ()
        response.vary = original_vary + ('Accept', 'Authorization')
        if request.authorization is not None:
            format = 'json'
        else:
            acceptable = request.accept.acceptable_offers(
                [
                    'text/html',
                    'application/ld+json',
                    'application/json',
                ])
            mime_type, q_value = acceptable[0] if acceptable else ('text/html', 0)
            format = mime_type.split('/', 1)[1]
            if format == 'ld+json':
                format = 'json'
    else:
        format = format.lower()
        if format not in ('html', 'json'):
            format = 'html'

    if format == 'json':
        return False

    request._transform_start = time.time()
    return True


def after_transform(request, response):
    end = time.time()
    duration = int((end - request._transform_start) * 1e6)
    stats = request._stats
    stats['render_count'] = stats.get('render_count', 0) + 1
    stats['render_time'] = stats.get('render_time', 0) + duration
    request._stats_html_attribute = True


# Rendering huge pages can make the node process memory usage explode.
# Ideally we would let the OS handle this with `ulimit` or by calling
# `resource.setrlimit()` from  a `subprocess.Popen(preexec_fn=...)`.
# Unfortunately Linux does not enforce RLIMIT_RSS.
# An alternative would be to use cgroups, but that makes per-process limits
# tricky to enforce (we would need to create one cgroup per process.)
# So we just manually check the resource usage after each transform.

rss_limit = 256 * (1024 ** 2)  # MB


def reload_process(process):
    return psutil.Process(process.pid).memory_info().rss > rss_limit

node_env = os.environ.copy()
node_env['NODE_PATH'] = ''

page_or_json = SubprocessTween(
    should_transform=should_transform,
    after_transform=after_transform,
    reload_process=reload_process,
    args=['node', resource_filename(__name__, 'static/build/renderer.js')],
    env=node_env,
)


debug_page_or_json = SubprocessTween(
    should_transform=should_transform,
    after_transform=after_transform,
    reload_process=reload_process,
    args=['node', resource_filename(__name__, 'static/server.js')],
    env=node_env,
)
